// ----- compiler.cpp

#include "stdafx.h"
#include "Quincy.h"
#include "compiler.h"
#include "ErrorLogDialog.h"

Compiler::Compiler() 
{
	m_bFoundDef = false;
	m_bCpp = false;
	m_bErrorCreated = false;
	m_bLinkCpp = false;
	m_nErrorCount = 0;
	m_bErrorsOpen = false;
	m_CompileStatus = idle;
	m_OldCompileStatus = idle;
	m_CompileAction = none;
	m_pdlgErrors = new CErrorLogDialog;
	m_nErrorLogWidth = 0;
}

Compiler::~Compiler()
{
	delete m_pdlgErrors;
}

void Compiler::ClearArrays()
{
	m_SourceFiles.RemoveAll();
	m_ObjectFiles.RemoveAll();
	m_LibraryFiles.RemoveAll();
	m_ResourceFiles.RemoveAll();
}

void Compiler::AddSourceFile(const CString& rstrSrc)
{
	m_SourceFiles.Add(rstrSrc);
}

void Compiler::AddObjectFile(const CString& rstrObj)
{
	m_ObjectFiles.Add(rstrObj);
}

void Compiler::AddResourceFile(const CString& rstrRc)
{
	m_ResourceFiles.Add(rstrRc);
}

void Compiler::BuildTarget(const CString& strTargetName, CompileAction action, bool bUseTD)
{
	BuildStdFiles();
	if (theApp.IsOnCDROM(strTargetName))	{
		std::string msg;
		msg += const_cast<CString&>(strTargetName).GetBuffer(0);
		msg += " is directed to a read-only medium (e.g. CD-ROM).",
		AfxMessageBox(msg.c_str(), MB_ICONSTOP);
		return;
	}
	m_CompileAction = action;
	m_strTargetName = strTargetName;
	ClearErrorLog();
	theApp.ShowStatusText(CString("Building Program"), &theApp.m_strOldStatus);
	AfxBeginThread(&Compiler::BldProg, 0);
}

UINT Compiler::BldProg(void*)
{
	theApp.GetCompiler().BuildTarget();
	return 0;
}

// --- put quotes around a file specification that has one or more spaces
//     so the file specification can be used on a command line
CString Compiler::Enquote(const CString& str) const
{
	if (str[0] == '"' || str.Find(' ') == -1)
		return str;		// file spec is already quoted or has no spaces
	CString qstr("\"" + str);
	// if the spec ends with backslash, add another backslash 
	if (qstr[qstr.GetLength()-1] == '\\')
		qstr += "\\";
	qstr += "\"";
	return qstr;
}

void Compiler::BuildTarget()
{
	m_strObjFiles.Empty();
	m_strLibFiles.Empty();
	m_strOldFilename.Empty();
	CompileAllSources();
	if (m_CompileStatus != aborting && ExitCode == 0)	{
		if (GatherObjects())	{
			if (theApp.IsLibraryTarget())	{
				m_CompileStatus = buildinglibrary;
				BuildLibraryTarget();
			}
			else if (theApp.IsDLLTarget())	{
				m_CompileStatus = buildinglibrary;
				BuildDLLTarget();
			}
			else	{
				m_CompileStatus = linking;
				BuildExeTarget();
			}
		}
	}
	if (ExitCode != 0)
		m_CompileAction = none;
	theApp.CloseConsoleHandles();
	m_CompileStatus = m_CompileStatus == aborting ? aborted : finished;
}

void Compiler::CompileAllSources()
{
	exitcode = 0;
	ExitCode = 0;
	m_bFoundDef = false;

	int nSize = m_SourceFiles.GetSize();
	// --- compile the programs listed in the source code list
	for (int n = 0; n < nSize && m_CompileStatus != aborting; n++)	{
		CString strFile = m_SourceFiles[n];
		if (strFile.IsEmpty())
			break;
		CompileOneSource(strFile);
		if (m_CompileStatus == aborting)
			break;
	}
}

void Compiler::CompileOneSource(const CString& strFile)
{
	CString strCmd;
	if (strFile.Right(4).CompareNoCase(".def") == 0)	{
		if (theApp.IsDLLTarget())	{
			BuildDefCommand(strCmd, strFile);
			m_bFoundDef = true;
		}
		else	{
			AfxMessageBox("Non-DLL project has .def file", MB_OK | MB_ICONQUESTION);
			ExitCode = 1;
			return;
		}
	}
	else	{
		bool bRc  = strFile.Right(3).CompareNoCase(".rc")  == 0;
		bool bRes = strFile.Right(4).CompareNoCase(".res") == 0;
		m_bCpp = strFile.Right(4).CompareNoCase(".cpp") == 0;
		m_bLinkCpp |= m_bCpp;

		// ---- make a copy of the filename without the path
		m_strFilename = theApp.GetFileName(strFile);

		// ---- make a copy of the source file path and name without the file extension
		CString strSrcPath = strFile.Left(strFile.GetLength()-((m_bCpp | bRes) ? 4 : (bRc ? 3 : 2)));

		// ---- make a copy of the object file path without the filename
		//      (object files go where the exe file goes.)
		int ndx = m_strTargetName.ReverseFind('\\');
		if (ndx != -1)
			m_strOPath = m_strTargetName.Left(ndx);
		else	{
			char path[MAX_PATH];
			_getcwd(path, MAX_PATH);
			m_strOPath = path;
		}

		// ---- make a copy of the object file name
		m_strOFile = MakeObjectFileName(strSrcPath);

		// if the object file spec has no path info, add it
		if (m_strOFile.Find("\\") == -1)
			m_strOFile = m_strOPath + "\\" + m_strOFile;

		if (bRc)	{
			// ------ compiling a resource script into res file
			CString strResFile = strSrcPath + ".res";
			remove(strResFile);
			BuildResourceScriptCommand(strCmd, strResFile);
		}
		else if (bRes)
			// --- converting resource .res file into object file
			BuildResFileCommand(strCmd);
		else	{
			remove(m_strOFile);
			BuildCompilerCommand(strCmd, strFile);
		}
	}
	m_CompileStatus = compiling;
	if (RunCompilerProgram(strCmd))
		ExitCode |= exitcode;
}

int Compiler::GatherObjects()
{
	int nSize = m_ObjectFiles.GetSize();
	int n = 0;
	if (nSize != 0)	{
		CString strFile;
		while (n < nSize)	{
			strFile = m_ObjectFiles[n++];
			if (strFile.IsEmpty())
				break;
			if (strFile.Find("\\") == -1)
				strFile = m_strOPath + "\\" + strFile;
			m_strObjFiles += Enquote(strFile) + " ";
		}
	}
	return n;
}

void Compiler::BuildLibraryTarget()
{
	CString strCmd;
	BuildLibCommand(strCmd);
	RunCompilerProgram(strCmd);
	ExitCode |= exitcode;
}

void Compiler::BuildDLLTarget()
{
	GatherLibraries();
	CString strCmd;
	BuildDLLCommand(strCmd);
	if (!strCmd.IsEmpty())
		if (!RunCompilerProgram(strCmd))
			exitcode = 1;
	ExitCode |= exitcode;
}

void Compiler::BuildExeTarget()
{
	GatherLibraries();
	CString strCmd;
	BuildLinkerCommand(strCmd);
	if (!RunCompilerProgram(strCmd))
		exitcode = 1;
	ExitCode |= exitcode;
	// --- if this is a compile and link of only one IDE translation unit
	//     and no user library files are in the project (no project, actually),
	//     no need to keep the object file around.
	if (!theApp.IsProjectFileLoaded())
		remove(m_strOFile);
}

void Compiler::GatherLibraries()
{
	int nSize = m_LibraryFiles.GetSize();
	CString strFile;
	for (int n = 0; n < nSize; n++)	{
		strFile = m_LibraryFiles[n];
		if (strFile.IsEmpty())
			break;
		m_strLibFiles += Enquote(strFile) + " ";
	}
}

bool Compiler::RunCompilerProgram(CString& strCmd, bool bIsRC)
{
	STARTUPINFO suinfo = {
		sizeof(STARTUPINFO),	// cb
		0, 0,					// reserved
		0,						// lpTitle
		0, 0,					// dwX, dwY
		0, 0,					// dwXSize, dwYSize
		0, 0,					// dwXCountChars, dwYCountChars
		0,						// dwFillAttribute
		STARTF_USESHOWWINDOW |	// dwFlags
		STARTF_USESTDHANDLES,
		SW_HIDE,				// wShowWindow
		0, 0,					// reserved
		0,						// hStdInput;
		theApp.m_hStdout,		// hStdOutput;
		theApp.m_hStderr,		// hStdError;
	};

	char* pCmd = strCmd.GetBuffer(0);
	bool rtn = CreateProcess(	0,
								pCmd,0,0,
								true, 
								bIsRC ? 0 : CREATE_NEW_CONSOLE,
								0,
								0,
								&suinfo,
								&m_CompileInformation) != false;
	if (rtn)	{
		while (WaitForSingleObject(m_CompileInformation.hProcess,100) == WAIT_TIMEOUT)
			if (m_CompileStatus == aborting)
				TerminateProcess(m_CompileInformation.hProcess,0);
		GetExitCodeProcess(m_CompileInformation.hProcess, &exitcode);
		CloseHandle(m_CompileInformation.hThread);
		m_CompileInformation.hThread = 0;
		CloseHandle(m_CompileInformation.hProcess);
		m_CompileInformation.hProcess = 0;
	}
	else	{
		AfxMessageBox(strCmd + "\nCannot execute compiler program", MB_ICONSTOP);
		m_CompileStatus = aborting;
	}
	return rtn;
}

void Compiler::ClearErrorLog()
{
	m_nErrorCount = 0;
	if (m_bErrorCreated == true)	{
		CListBox* pList = static_cast<CListBox*>(m_pdlgErrors->GetDlgItem(IDC_ERRORLIST));
		ASSERT(pList != 0);
		pList->ResetContent();
		m_nErrorLogWidth = 0;
	}
}

// --- called from the idle loop while compiling is going on
void Compiler::CollectErrorMessages()
{
	if (!m_bErrorsOpen)
		OpenErrorMessagesFile();
	if (m_bErrorsOpen)	{
		char buf[201];
		while (m_ErrorCStdioFile.ReadString(buf, 200) != 0)	{
			if (m_nErrorCount < maxerrors)	{
				int hold = m_nErrorCount;
				ProcessErrorMessage(buf);
				if (hold != m_nErrorCount)	{
					// an error message was created, add it to the error log display
					ASSERT(m_pdlgErrors != 0);
					if (m_bErrorCreated != true)	{
						m_pdlgErrors->Create(IDD_ERRORLOG);
						m_bErrorCreated = true;
					}
					m_pdlgErrors->ShowWindow(SW_SHOW);
					CListBox* pList = static_cast<CListBox*>(m_pdlgErrors->GetDlgItem(IDC_ERRORLIST));
					ASSERT(pList != 0);
					CDC* pDC = pList->GetDC();
					CString& strErr = ErrorLog[hold].m_strErrorMsg; 
					pList->AddString(strErr);
					CSize size = pDC->GetTextExtent(strErr);
					m_nErrorLogWidth = max(m_nErrorLogWidth, size.cx);
					pList->SendMessage(LB_SETHORIZONTALEXTENT, m_nErrorLogWidth, 0);
				}
			}
		}
	}
}

void Compiler::OpenErrorMessagesFile()
{
	CString strErr(theApp.WorkingFile("quincy.err"));
	if (_access(strErr, 0) != 0)
		return;
	m_bErrorsOpen = (m_ErrorCStdioFile.Open(strErr, CFile::modeRead | CFile::shareDenyNone) != 0);
//	ShowErrorLog();
}

void Compiler::CloseErrorMessagesFile()
{
	CollectErrorMessages();
	if (m_bErrorsOpen)
		m_ErrorCStdioFile.Close();
	remove(theApp.WorkingFile("quincy.out"));
	m_bErrorsOpen = false;
}

void Compiler::CompileStep()
{
	switch (m_CompileStatus)	{
		case preprocessing:
		case compiling:
			theApp.ShowStatusText(CString("Compiling " + m_strFilename));
			break;
		case linking:
			theApp.ShowStatusText(CString("Linking " + m_strTargetName));
			break;
		case buildinglibrary:
			theApp.ShowStatusText(CString("Building " + m_strTargetName));
			break;
		case finished:
		case aborted:
			theApp.ShowStatusText(theApp.m_strOldStatus);
			if (m_CompileStatus == aborted)
				AfxMessageBox("User terminated compile", MB_ICONEXCLAMATION);
			m_CompileStatus = idle;
			if (m_CompileAction != none)
				theApp.RunApplicationProgram(m_strTargetName, m_CompileAction == step, m_CompileAction == turbodebugger);
			m_CompileAction = none;
			break;
		default:
			break;
	}
}

bool Compiler::OnIdle()
{
	if (m_CompileStatus != m_OldCompileStatus || (m_CompileStatus == compiling && m_strFilename != m_strOldFilename))	{
		CompileStep();
		m_OldCompileStatus = m_CompileStatus;
		m_strOldFilename = m_strFilename;
	}
	if (m_CompileStatus != idle)	{
		CollectErrorMessages();
		if (theApp.m_pMainWnd != 0)	{
			theApp.m_pMainWnd->KillTimer(24);
			theApp.m_pMainWnd->SetTimer(24, 250, 0);
		}
		return false;
	}

	if (m_bErrorsOpen)
		CloseErrorMessagesFile();

	return true;
}

// ----- called from CQuincyApp to show the error log if user has closed it
void Compiler::ShowErrorLog() 
{
	ASSERT(m_pdlgErrors != 0);
	if (m_bErrorCreated != true)	{
		m_pdlgErrors->Create(IDD_ERRORLOG);
		m_bErrorCreated = true;
	}
	CListBox* pList = static_cast<CListBox*>(m_pdlgErrors->GetDlgItem(IDC_ERRORLIST));
	ASSERT(pList != 0);
	pList->ResetContent();

	CDC* pDC = pList->GetDC();
	m_nErrorLogWidth = 0;

	for (UINT i = 0; i < m_nErrorCount; i++)	{
		CString& strErr = ErrorLog[i].m_strErrorMsg; 
		pList->AddString(strErr);
		CSize size = pDC->GetTextExtent(strErr);
		m_nErrorLogWidth = max(m_nErrorLogWidth, size.cx);
	}
	pList->SendMessage(LB_SETHORIZONTALEXTENT, m_nErrorLogWidth, 0);
	m_pdlgErrors->ShowWindow(SW_SHOW);
}

////////////////////////////////////////////////////////////////////////
//
// Specific compiler specializations of the abstract base Compiler class
//
////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////
//
//                GCCCompiler member functions
//
///////////////////////////////////////////////////////////////

void GCCCompiler::CompileAllSources()
{
	m_strPrevLine.Empty();
	Compiler::CompileAllSources();
}

void GCCCompiler::CompileOneSource(const CString& strFile)
{
	const_cast<CString&>(strFile).MakeLower();	// gcc 2.95.1  bug
	Compiler::CompileOneSource(strFile);
}

void GCCCompiler::AddLibraryFile(const CString& rstrLib)
{
	if (rstrLib.Left(3).CompareNoCase("lib") == 0)	{
		CString strLib("-l");
		strLib += rstrLib.Mid(3, rstrLib.GetLength()-5);
		m_LibraryFiles.Add(strLib);
	}
	else
		m_LibraryFiles.Add(rstrLib);
}

void GCCCompiler::BuildStdFiles()
{
	// GCC sends its warnings and error messages to stderr
	m_strErrorFile = theApp.WorkingFile("quincy.err");
	theApp.BuildStderr(m_strErrorFile);
	// repository for stdout messages, which will be discarded
	theApp.BuildStdout(theApp.WorkingFile("quincy.out"));
}

void GCCCompiler::BuildResourceScriptCommand(CString& strCmd, const CString& strResFile)
{
	char path[MAX_PATH];
	_getcwd(path, MAX_PATH);

	strCmd = theApp.LCCPath() + "lrc.exe";
	strCmd += " /o";
	strCmd += Enquote(strResFile);
	strCmd += " /i";
	strCmd += Enquote(path);
	strCmd += " /i";
	strCmd += Enquote(theApp.RIncludePath());
	strCmd += " ";
	strCmd += Enquote(m_strFilename);
}

void GCCCompiler::BuildResFileCommand(CString& strCmd)
{
	strCmd = theApp.ToolsPath() + "windres.exe";
	strCmd += " -i ";
	strCmd += Enquote(m_strFilename);
	strCmd += " -o ";
	strCmd += Enquote(m_strOFile);
}

void GCCCompiler::BuildCompilerCommand(CString& strCmd, const CString& strFile)
{
	strCmd = theApp.ToolsPath() + "gcc";
	if (theApp.IsAddingDebugInfo())
		strCmd += " -g";
	if (!theApp.IsAddingDebugInfo() && theApp.OptimizeLevel())	{
			strCmd += " -O";
			char oc = '0' + theApp.OptimizeLevel();
			strCmd += oc;
	}
	if (m_bCpp)	{
		if (!theApp.IncludeRTTI())
			strCmd += " -fno-rtti";
		if (!theApp.IncludeExceptions())
			strCmd += " -fno-exceptions";
	}

	if (theApp.IsStdLibInstalled() && theApp.UseStdLib())
		strCmd += " -fhonor-std";

	else if (theApp.StrictANSI())
		strCmd += " -pedantic-errors";

	int i;
	for (i = 0; i < theApp.GetIncludeArray().GetSize(); i++)	{
		strCmd += " -I";
		strCmd += Enquote(theApp.GetIncludeArray()[i]);
	}

	if (theApp.IsStdLibInstalled() && theApp.UseStdLib())	{
		strCmd += " -I";
		strCmd += Enquote(theApp.StdIncludePath());
	}

	if (theApp.IsGUITarget())
			strCmd += " -mwindows";

	if (theApp.IsDLLTarget())
			strCmd += " -mdll";

	for (i = 0; i < theApp.GetDefinesArray().GetSize(); i++)	{
		strCmd += " -D";
		strCmd += theApp.GetDefinesArray()[i];
	}

	strCmd += " -o ";
	strCmd += Enquote(m_strOFile);
	strCmd += " -c ";

	strCmd += theApp.CmdLineOptions() + " ";

	strCmd += Enquote(strFile);
}

void GCCCompiler::BuildLibCommand(CString& strCmd)
{
	strCmd = theApp.ToolsPath() + "ar";
	strCmd += " rc ";
	strCmd += Enquote(m_strTargetName);
	strCmd += " ";
	strCmd += m_strObjFiles;
}

void GCCCompiler::BuildDLLCommand(CString& strCmd)
{
	if (!m_bFoundDef)	{
		AfxMessageBox("DLL project has no .def file", MB_OK | MB_ICONSTOP);
		return;
	}
	// ------ build dll
	GatherLibraries();

	strCmd = Enquote(theApp.ToolsPath() + "dllwrap.exe") + " -o ";
	strCmd += Enquote(theApp.GetFileName(m_strTargetName));		// name of the dll
	strCmd += " --def ";
	strCmd += Enquote(m_strDefname);		// name of the def

	GetLibraryPaths(strCmd);

	strCmd += m_strObjFiles;
	strCmd += " ";
	strCmd += m_strLibFiles;

	GetGUILibraries(strCmd);

}

void GCCCompiler::BuildDefCommand(CString& strCmd, const CString& strFile)
{
	int ndx = m_strTargetName.ReverseFind('\\');
	// --- build the path to the import library
	CString strImportPath = m_strTargetName.Left(ndx+1);
	CString strImportName = m_strTargetName.Right(m_strTargetName.GetLength() - (ndx+1));
	CString strImportLib = strImportPath + "lib" + strImportName.Left(strImportName.GetLength() - 3) + "a";

	m_strDefname = strImportPath + strFile;

	strCmd = theApp.ToolsPath() + "dlltool.exe";
	strCmd += " --dllname ";
	strCmd += Enquote(theApp.GetFileName(m_strTargetName));		// name of the dll
	strCmd += " --def ";
	strCmd += Enquote(m_strDefname);		// name of the def
	strCmd += " --output-lib ";
	strCmd += Enquote(strImportLib);		// name of the import lib
}

void GCCCompiler::GetLibraryPaths(CString& strCmd)
{
	if (theApp.IsStdLibInstalled() && theApp.UseStdLib())	{
		strCmd += " -L";
		strCmd += Enquote(theApp.StdLibPath());
		strCmd += " ";
	}
	int i;
	for (i = 0; i < 4; i++)	{
		strCmd += " -L";
		strCmd += Enquote(theApp.LibPath()[i]);
	}

	for (i = 0; i < theApp.GetLibraryArray().GetSize(); i++)	{
		strCmd += " -L";
		strCmd += Enquote(theApp.GetLibraryArray()[i]);
	}
	strCmd += " ";
}

void GCCCompiler::GetGUILibraries(CString& strCmd)
{
	strCmd += " -luser32 -lgdi32 -lcomdlg32 -lwinmm -lcomctl32 -lctl3d32 ";
}

void GCCCompiler::BuildLinkerCommand(CString& strCmd)
{
	exitcode = 0;

	strCmd = theApp.ToolsPath() + "gcc ";

	if (theApp.IsGUITarget())
		strCmd += "-mwindows ";
	strCmd += "-o ";
	strCmd += Enquote(m_strTargetName) + " ";

	if (!theApp.IsAddingDebugInfo())
		strCmd += "-s ";

	if (theApp.IsGUITarget() && theApp.IsAddingDebugInfo())
		strCmd += Enquote(theApp.QuincyBinPath() + "\\trace.o") + " ";

	GetLibraryPaths(strCmd);

	strCmd += m_strObjFiles;
	strCmd += " ";
	strCmd += m_strLibFiles;

	if (m_bLinkCpp)
		strCmd += " -lstdc++";

	if (theApp.IsGUITarget())
		GetGUILibraries(strCmd);

}

void GCCCompiler::ProcessErrorMessage(char* buf)
{
	// --- messages from gnu-mingw32
	if (strncmp(buf, "GNU", 3) == 0)
		return;
	char *cp1 = buf;
	char *cp2 = buf;

	if (*cp1 == '/' && *(cp1+1) == '/' && *(cp1+3) == '/')	{
		// --- convert //C/ to C:/
		*cp1++ = *(cp1+2);	// drive letter
		*cp1++ = ':';
		cp2 = cp1++;
	}

	// ---- strip double slashes and backslashes from the error message
	while (*cp1)	{
		if (*cp1 == '\\' && *(cp1+1) == '\\')
			cp1++;
		else if (*cp1 == '/' && *(cp1+1) == '/')
			cp1++;
		*cp2++ = *cp1++;
	}
	*cp2 = '\0';

	CString strLn(buf);

	int nLinkerror = 0;
	int offset = strLn.Find(".c:");
	if (offset == -1)	{
		offset = strLn.Find(".cpp:");
		if (offset == -1)	{
			offset = strLn.Find(".h:");
			if (offset == -1)	{
				offset = strLn.Find(".i:");
				if (offset == -1)	{
					offset = strLn.Find(".ii:");
					if (offset == -1)	{
						offset = strLn.Find(".cc:");
						if (offset == -1)	{
							offset = strLn.Find(".rc:");
							if (offset != -1)	{
								nLinkerror = 4;	// .rc compiler error
								offset++;
							}
						}
						else	{
							nLinkerror = 3;	// .cc compiler error
							offset++;
						}
					}
					else
						nLinkerror = 2;	// .cpp link error
				}
				else
					nLinkerror = 1;		// .c link error
			}
		}
		else
			offset += 2;	// .cpp compiler error
	}

	ErrorLog[m_nErrorCount].m_strFileSpec = "";
	ErrorLog[m_nErrorCount].m_nLineNumber = 0;

	CString strer;

	if (offset == -1)
		strer = strLn;	// unparsed error message, display as is

	else if (nLinkerror != 0 && nLinkerror != 3 && nLinkerror != 4)	{
		// ---- linker error message, scan to beginning of .i or .ii file name
		char ch;
		int foffset = offset > 0 ? offset-1 : 0;
		while (foffset >= 0 && (ch = strLn[foffset]) == '_' || isalpha(ch) || isdigit(ch))
			--foffset;
		strer = strLn.Mid(foffset+1, offset-foffset);
		strer += "c";
		if (nLinkerror == 2)
			strer += "pp";
		ErrorLog[m_nErrorCount].m_strFileSpec = strer;
		strer += strLn.Mid(offset+1+nLinkerror);
	}
	else	{
		// ---- compiler warning/error message found
		offset += 3;
		if (strLn[offset] == ' ' && nLinkerror != 3 && nLinkerror != 4)
			// ---- no line number, might be a library link error
			strer = strLn;	// display as is
		else	{
			if (nLinkerror != 3)	{
				// ---- retrieve the source code line number for this error
				UINT nLineNo = atoi(strLn.GetBuffer(0)+offset);
				ErrorLog[m_nErrorCount].m_nLineNumber = nLineNo;
				while (isdigit(strLn[offset]))
					offset++;
				if (strLn.Left(offset) == m_strPrevLine)
					return;
				m_strPrevLine = strLn.Left(offset);
			}
			// ---- parse out file name and error message
			char ch;
			int foffset = 0;
			--offset;
			while ((ch = strLn[offset]) != '\\' && ch != '/' && ch != ' ')	{
				if (ch == ':')	{
					if (foffset == 0)
						foffset = offset;
					else if (nLinkerror == 3)
						break;
				}
				if (offset-- == 0)
					break;
			}
			strer = (strLn.Right(strLn.GetLength() - (offset+1)));

			if (nLinkerror != 3)	{
				// ------ parse out file spec
				CString strFileSpec = strLn.Left(foffset);
				--foffset;
				while (strFileSpec[foffset] != ' ')
					if (foffset-- == 0)
						break;
				ErrorLog[m_nErrorCount].m_strFileSpec = 
					strFileSpec.Right(strFileSpec.GetLength()-(foffset+1));
			}
		}
	}

	// ------ put error entry into error log array
	ErrorLog[m_nErrorCount].m_strErrorMsg = strer.Left(strer.GetLength()-1);
	m_nErrorCount++;
}

///////////////////////////////////////////////////////////////
//
//                BCCCompiler member functions
//
///////////////////////////////////////////////////////////////

void BCCCompiler::AddLibraryFile(const CString& rstrLib)
{
	m_LibraryFiles.Add(rstrLib);
}

void BCCCompiler::BuildStdFiles()
{
	// BCC sends its warnings and error messages to stdout
	m_strErrorFile = theApp.WorkingFile("quincy.err");
	theApp.BuildStdout(m_strErrorFile);
	// repository for stderr messages, which will be discarded
	theApp.BuildStderr(theApp.WorkingFile("quincy.out"));
}

void BCCCompiler::BuildResourceScriptCommand(CString& strCmd, const CString& strResFile)
{
	strCmd = theApp.ToolsPath() + "brcc32 ";
	strCmd += " -fo ";
	strCmd += Enquote(strResFile);
	strCmd += " -i ";
	strCmd += Enquote(theApp.RIncludePath());
	strCmd += " ";
	strCmd += Enquote(m_strFilename);
}

void BCCCompiler::BuildResFileCommand(CString& strCmd)
{
	// null function should never be called. bcc does not compile .res into .obj
	ASSERT(0);
}

void BCCCompiler::BuildCompilerCommand(CString& strCmd, const CString& strFile)
{
	strCmd = theApp.ToolsPath() + "bcc32";
	if (theApp.IsAddingDebugInfo())
		strCmd += " -v ";
	else
		strCmd += " -v- ";

	if (!theApp.IsAddingDebugInfo() && theApp.OptimizeLevel())	{
			strCmd += " -O";
			char oc = '0' + theApp.OptimizeLevel();
			strCmd += oc;
	}
	if (m_bCpp)	{
		if (!theApp.IncludeRTTI())
			strCmd += " -RT-";
		if (!theApp.IncludeExceptions())
			strCmd += " -x-";
	}

	int i;
	for (i = 0; i < theApp.GetIncludeArray().GetSize(); i++)	{
		strCmd += " -I";
		strCmd += Enquote(theApp.GetIncludeArray()[i]);
	}

	if (theApp.IsGUITarget())
			strCmd += " -W";

	if (theApp.IsDLLTarget())
			strCmd += " -WD";

	for (i = 0; i < theApp.GetDefinesArray().GetSize(); i++)	{
		strCmd += " -D";
		strCmd += theApp.GetDefinesArray()[i];
	}

	strCmd += " -n";
	strCmd += Enquote(m_strOPath);

	strCmd += " -c ";

	strCmd += theApp.CmdLineOptions() + " ";

	strCmd += Enquote(strFile);
}

void BCCCompiler::BuildLibCommand(CString& strCmd)
{
	strCmd = theApp.ToolsPath() + "tlib /a ";
	strCmd += Enquote(m_strTargetName);
	strCmd += " ";
	strCmd += m_strObjFiles;
}

void BCCCompiler::BuildDLLCommand(CString& strCmd)
{
	strCmd = theApp.ToolsPath() + "bcc32 -WD ";
	if (!m_bFoundDef)
		strCmd += "-lGi ";
	strCmd += "-e";
	strCmd += Enquote(m_strTargetName) + " ";
	GetLibraryPaths(strCmd);
	strCmd += " ";
	strCmd += m_strObjFiles;
	strCmd += " ";
	strCmd += m_strLibFiles;
}

void BCCCompiler::BuildDefCommand(CString& strCmd, const CString& strFile)
{
	strCmd = theApp.ToolsPath() + "implib ";

	int ndx = m_strTargetName.ReverseFind('\\');
	// --- build the path to the import library
	CString strImportPath = m_strTargetName.Left(ndx+1);
	CString strImportName = m_strTargetName.Right(m_strTargetName.GetLength() - (ndx+1));
	CString strImportLib = m_strTargetName.Left(m_strTargetName.GetLength() - 3) + "lib";
	m_strDefname = strImportPath + strFile;

	strCmd += strImportLib + " " + m_strDefname;
}

void BCCCompiler::GetLibraryPaths(CString& strCmd)
{
	strCmd += "-L";
	strCmd += Enquote(theApp.LibPath()[0]);
	for (int i = 0; i < theApp.GetLibraryArray().GetSize(); i++)	{
		strCmd += " -L";
		strCmd += Enquote(theApp.GetLibraryArray()[i]);
	}
}

void BCCCompiler::BuildLinkerCommand(CString& strCmd)
{
	exitcode = 0;

	int nSize = m_ResourceFiles.GetSize();

	if (nSize > 0)	{
		// --- Win32 exe with resources
		// ----- build the bcc linker command
		strCmd = theApp.ToolsPath() + "ilink32 /aa -x -Gn ";
		if (theApp.IsAddingDebugInfo())
			strCmd += "-v ";
		GetLibraryPaths(strCmd);
		strCmd += " " + theApp.StartupCode() + "c0w32.obj ";
		if (theApp.IsAddingDebugInfo())
			strCmd += Enquote(theApp.QuincyBinPath() + "\\trace.obj") + " ";
		strCmd += m_strObjFiles;
		strCmd += " ";
		strCmd += m_strLibFiles;
		strCmd += "," + Enquote(m_strTargetName) + ",,";
		strCmd += "cw32.lib import32.lib,,"; //// might need more of these -- later
		for (int i = 0; i < nSize; i++)
			strCmd += Enquote(m_ResourceFiles[i]) + " ";
	}
	else	{
		// ----- build the bcc linker command
		strCmd = theApp.ToolsPath() + "bcc32 ";

		if (theApp.IsAddingDebugInfo())
			strCmd += "-v ";
		else
			strCmd += "-v- ";

		if (theApp.IsGUITarget())
			strCmd += "-W ";

		if (theApp.IsDLLTarget())
			strCmd += "-WD ";

		GetLibraryPaths(strCmd);

		strCmd += " -e" + Enquote(m_strTargetName);

		strCmd += " ";
		if (theApp.IsGUITarget() && theApp.IsAddingDebugInfo())
			strCmd += Enquote(theApp.QuincyBinPath() + "\\trace.obj") + " ";
		strCmd += m_strObjFiles;
		strCmd += " ";
		strCmd += m_strLibFiles;
	}
}

void BCCCompiler::ProcessErrorMessage(char* buf)
{
	// --- messages from bcc 5.5
	CString strer;
	ErrorLog[m_nErrorCount].m_strFileSpec = "";
	ErrorLog[m_nErrorCount].m_nLineNumber = 0;
	strer = buf;
	// "Error:" is linker error, no line number
	if (strncmp(buf, "Error:", 6) != 0)	{
		bool iserror   = strncmp(buf, "Error ",   6) == 0;
		bool iswarning = strncmp(buf, "Warning ", 8) == 0;
		if (!iserror && !iswarning)
			return;		// just BCC command line information, don't display
		char* cp = buf + (iswarning ? 14 : 12);

		// parse out the file specification
		CString str;
		while (*cp && *cp != ' ')
			str += *cp++;
		ErrorLog[m_nErrorCount].m_strFileSpec = str;
		// parse out the line number
		ErrorLog[m_nErrorCount].m_nLineNumber = atoi(++cp);
	}
	// ------ put error entry into error log array
	ErrorLog[m_nErrorCount].m_strErrorMsg = strer.Left(strer.GetLength()-1);
	m_nErrorCount++;
}

void BCCCompiler::BuildExeTarget()
{
	Compiler::BuildExeTarget();
	if (!theApp.IsAddingDebugInfo())
		// ---- delete the .tds file that Borland always builds
		remove(m_strTargetName.Left(m_strTargetName.GetLength()-3) + "tds");
}
