// ------ parser.cpp

#include "stdafx.h"
#include "quincy.h"
#include "parser.h"
#include "stabs.h"

void ParserError::DisplayMessage(const char* pExpression)
{
	if (!m_strErrorMsg.IsEmpty())	{
		CString strMsg;
		strMsg.Format("Watch/Examine: %s\n%s\n", m_strErrorMsg.GetBuffer(0), pExpression);
		AfxMessageBox(strMsg, MB_ICONEXCLAMATION);
	}
}

Item::Item()
{
	NullItem();
}

Item::Item(Item& rItem)
{
	NullItem();
	Copy(rItem);
}

Item::Item(int nConstant)
{
	NullItem();
	m_Type = Item::Int;
	m_Datum.m_short = nConstant;
	m_nMemaddr = 0;
}

Item::Item(const CString& strIdentifier)
{
	NullItem();

	CStab* pCStabs = theApp.GetDebugger()->GetStabs();
	ASSERT(pCStabs != 0);

	const Symbol* pSym = pCStabs->FindSymbol(strIdentifier, theApp.GetDebugger()->CurrentPC());
	if (pSym == 0)	{
		if (strIdentifier != "this")	{
			// --- can't find the identifier, try this->identifier
			try	{
				Item thisItem("this");
				if (thisItem.GetItemType() != Item::Structure)
					throw UnknownSymbol();
				thisItem.DereferencePointer();
				thisItem.GetMember(strIdentifier);
				Copy(thisItem);
				return;
			}
			catch (ParserError)	{
				throw UnknownSymbol();
			}
		}
		else
			throw UnknownSymbol();
	}

	m_StabType = pSym->m_TypeIndex;
	m_pType = pCStabs->GetType(m_StabType);
	if (m_pType == 0)
		throw UnknownType();

	m_nMemaddr = pSym->address;
	if (pSym->storageclass == Symbol::stack)
		m_nMemaddr += theApp.GetDebugger()->Context().Ebp;

	if (pSym->storageclass == Symbol::function)	{
		m_strTypeDisp = m_pType->GetName();
		m_strTypeDisp += "()";
	}
	else
		InsertTypeData();
}

Item::~Item()
{
	delete [] m_pDimensions;
}

void Item::InsertTypeData()
{
	CStab* pCStabs = theApp.GetDebugger()->GetStabs();
	ASSERT(pCStabs != 0);

	for (;;)	{
	
//		ASSERT(m_pType != 0);
		if (m_pType == 0)
			throw UnknownSymbol();

		Reference* pRef = dynamic_cast<Reference*>(m_pType);

		if (pRef != 0)	{
			m_nReferences++;
			ReadProcessMemory(theApp.GetDebugger()->ProcessHandle(), (void*) m_nMemaddr, &m_nMemaddr, sizeof(MemoryAddress), 0);
			m_StabType = pRef->m_TypeIndex;
			m_pType = pCStabs->GetType(m_StabType);
			continue;
		}

		Pointer* pPtr = dynamic_cast<Pointer*>(m_pType);

		if (pPtr != 0)	{
			m_StabType = pPtr->m_TypeIndex;
			m_pType = pCStabs->GetType(m_StabType);
			m_nIndirection++;
			continue;
		}

		Array* pArr = dynamic_cast<Array*>(m_pType);

		if (pArr != 0)	{
			m_Datum.m_ptr = (char*) m_nMemaddr;
			short int* pd = new short int[m_nDimensions+1];
			for (int i = 0; i < m_nDimensions; i++)
				pd[i] = m_pDimensions[i];
			pd[i] = pArr->m_nElements;
			delete m_pDimensions;
			m_nDimensions++;
			m_pDimensions = pd;
			m_StabType = pArr->m_TypeIndex;
			m_pType = pCStabs->GetType(m_StabType);
			continue;
		}

		Typedef* pTyp = dynamic_cast<Typedef*>(m_pType);

		if (pTyp != 0)	{
			m_StabType = pTyp->m_TypeIndex;
			m_pType = pCStabs->GetType(m_StabType);
			continue;
		}

		break;
	}

	m_nWidth = m_pType->GetWidth();

	SetType();
	m_strTypeDisp = 
		CString(m_Type == Structure ? "struct " : m_Type == Enum ? "enum "  : "")
			+ m_pType->GetName();


	if (isLvalue())
		GetFromMemory();
}

void Item::NullItem()
{
	m_Datum.m_double = 0;
	m_nIndirection = 0;
	m_Type = Void;
	m_nWidth = 0;
	m_nDimensions = 0;
	m_pDimensions = 0;
	m_nMemaddr = 0;
	m_pType = 0;
	m_nReferences = 0;
}

void Item::Copy(Item& rItem)
{
	m_Type = rItem.m_Type;
	m_Datum = rItem.m_Datum;
	m_strTypeDisp = rItem.m_strTypeDisp;
	m_nIndirection = rItem.m_nIndirection;
	m_nDimensions = rItem.m_nDimensions;
	m_nWidth = rItem.m_nWidth;
	m_nMemaddr = rItem.m_nMemaddr;
	m_StabType = rItem.m_StabType;
	m_pType = rItem.m_pType;

	delete [] m_pDimensions;
	m_pDimensions = 0;
	if (m_nDimensions)	{
		m_pDimensions = new short int[m_nDimensions];
		for (int i = 0; i < m_nDimensions; i++)
			*(m_pDimensions + i) = *(rItem.m_pDimensions + i);
	}
}

void Item::AddressOf()
{
	if (!isLvalue())
		throw LValueError();
	// --- convert to address of item
	m_Datum.m_ptr = (char*) m_nMemaddr;
	m_nMemaddr = 0;
	m_nIndirection++;
}

Item& Item::operator=(Item& rItem)
{
	Copy(rItem);
	return *this;
}

short int Item::Width()
{
	int width = 1;
	if (isArray())
		for (int i = 1; i < m_nDimensions; i++)
			width *= *(m_pDimensions + i);
	return width * m_nWidth;
}

void Item::SetType()
{
	if (dynamic_cast<Enumerated*>(m_pType))
		m_Type = Enum;
	else if (m_StabType.hdrno != 0)
		m_Type = Structure;
	else	{
		const char* cp = m_pType->GetName();
		ASSERT(cp != 0);
		if (strcmp(cp, "bool") == 0)
			m_Type = Bool;
		else if (strcmp(cp, "void") == 0)
			m_Type = Void;
		else	{
			switch (m_StabType.typeno)	{
				case 2:
				case 10:
					m_Type = Char;
					break;
				case 11:
					m_Type = UnsignedChar;
					break;
				case 1:
					m_Type = Int;
					break;
				case 3:
					m_Type = Long;
					break;
				case 4:
					m_Type = UnsignedInt;
					break;
				case 5:
					m_Type = UnsignedLong;
					break;
				case 8:
					m_Type = Short;
					break;
				case 9:
					m_Type = UnsignedShort;
					break;
				case 12:
					m_Type = Float;
					break;
				case 13:
					m_Type = Double;
					break;
				case 14:
					m_Type = LongDouble;
					break;
				case 15:
				case 16:
				case 17:
				case 18:
					throw UnsupportedTypeError();
					break;
				default:
					m_Type = Structure;
					break;
			}
		}
	}
}

void Item::GetFromMemory()
{
	// ----- read the variable's data from debugged application's memory
	TypeUnion tu;

	int width = isPointer() ? sizeof(void*) : m_nWidth;
	ReadProcessMemory(theApp.GetDebugger()->ProcessHandle(), (void*) m_nMemaddr, &tu, width, 0);

	// ------ translate the variable data into an Item
	if (isPointer())
		m_Datum.m_ptr = (char*) tu.T15;
	else if (m_Type == Enum || m_Type == Bool)
		m_Datum.m_char = tu.T2;
	else	{
		switch (m_StabType.typeno)	{
			case 2:
			case 10:
			case 11:
				m_Datum.m_char = tu.T2;
				break;
			case 1:
				m_Datum.m_int = tu.T1;
				break;
			case 3:
				m_Datum.m_long = tu.T3;
				break;
			case 4:
				m_Datum.m_unsigned = tu.T4;
				break;
			case 5:
				m_Datum.m_unsignedlong = tu.T5;
				break;
			case 8:
				m_Datum.m_short = tu.T8;
				break;
			case 9:
				m_Datum.m_unsignedshort = tu.T9;
				break;
			case 12:
				m_Datum.m_float = tu.T12;
				break;
			case 13:
				m_Datum.m_double = tu.T13;
				break;
			case 14:
				m_Datum.m_longdouble = tu.T14;
				break;
		}
	}
}

void Item::DereferencePointer()
{
	if (!isPointerOrArray())
		throw PtrArrayError();
	NoVoidPointer();

	if (isArray())	{
		for (int i = 0; i < m_nDimensions-1; i++)
			*(m_pDimensions+i) = *(m_pDimensions+i+1);
		*(m_pDimensions+i) = 0;
		if (--m_nDimensions == 0)	{
			m_nMemaddr = (MemoryAddress) m_Datum.m_ptr;
			m_Datum.m_ptr = 0;
		}
	}

	if (!isArray() && isPointer())	{
		--m_nIndirection;
		m_nMemaddr = (MemoryAddress) m_Datum.m_ptr;
		m_Datum.m_ptr = 0;
	}
	if (isLvalue())
		GetFromMemory();
}

void Item::GetMember(const CString& rstrIdentifier)
{
	CStab* pCStabs = theApp.GetDebugger()->GetStabs();
	ASSERT(pCStabs != 0);
	ASSERT(m_pType != 0);

	Struct* pStruct = dynamic_cast<Struct*>(m_pType);
	if (pStruct != 0)	{
		std::vector<Member>::const_iterator iter = pStruct->members.begin();
		while (iter != pStruct->members.end())	{
			const Member& rMem = *iter;
			if (rstrIdentifier == rMem.m_pName)	{
				m_nMemaddr += rMem.m_nOffset;
				m_StabType = rMem.m_TypeIndex;
				m_pType = pCStabs->GetType(m_StabType);
				ASSERT(m_pType != 0);
				InsertTypeData();
				return;
			}
			iter++;
		}
	}
	throw UnknownSymbol();
}

void Item::NoVoidPointer()
{
    if (isPointerOrArray() && m_Type == Void)
		throw VoidPointerError();
}
void Item::IntegersOnly(Item& rItem)
{
    if (isPointerOrArray() || rItem.isPointerOrArray())
		throw IntegerError();
	NeedInteger();
	rItem.NeedInteger();
}
void Item::operator+=(Item& rItem)
{
	NoVoidPointer();
	rItem.NoVoidPointer();

    if (isPointerOrArray() && rItem.isPointerOrArray())
		throw PointerAdditionError();
	if (isPointerOrArray())	{
		rItem.NeedInteger();
		m_Datum.m_ptr += (rItem.m_Datum.m_int * Width());
	}
    else if (rItem.isPointerOrArray())    {
		NeedInteger();
		m_Datum.m_ptr = 
			rItem.m_Datum.m_ptr += (m_Datum.m_int * rItem.Width());
	}
	else	{
		NeedInteger();
		rItem.NeedInteger();
		m_Datum.m_int += rItem.m_Datum.m_int;
	}
}

void Item::operator-=(Item& rItem)
{
	NoVoidPointer();
	rItem.NoVoidPointer();

	if (isPointerOrArray() && rItem.isPointerOrArray())	{
		if (m_nWidth != rItem.m_nWidth)
			throw IncompatiblePointerSubtractionError();
		m_Datum.m_int = (m_Datum.m_ptr - rItem.m_Datum.m_ptr) / Width();
	}
	else if (isPointerOrArray())	{
		rItem.NeedInteger();
		m_Datum.m_ptr -= rItem.m_Datum.m_int * Width();
	}
	else if (rItem.isPointerOrArray())
		throw PointerSubtractionError();
	else	{
		NeedInteger();
		rItem.NeedInteger();
		m_Datum.m_int -= rItem.m_Datum.m_int;
	}
}

void Item::operator*=(Item& rItem)
{
	IntegersOnly(rItem);
	m_Datum.m_int *= rItem.m_Datum.m_int;
}

void Item::operator/=(Item& rItem)
{
	IntegersOnly(rItem);
	m_Datum.m_int /= rItem.m_Datum.m_int;
}

void Item::BuildValueDisplay(CString& strVarValue)
{
	if (m_strTypeDisp.IsEmpty())	{
		strVarValue = "??";
		return;
	}
	strVarValue = m_strTypeDisp;
	CString strFmt;
	if (m_Type == Char && isPointerOrArray())	{
		char buf[51];
		buf[50] = '\0';
		ReadProcessMemory(theApp.GetDebugger()->ProcessHandle(), (void*) m_Datum.m_ptr, buf, 50, 0);
		strFmt.Format(isArray() ? " = \"%s\"" : " -> \"%s\"", buf);
	}
	else if (m_nIndirection || isArray())
		strFmt.Format(" = %p", m_Datum.m_ptr);
	else	{
		switch (m_Type)	{
			case Bool:
				strFmt = m_Datum.m_char ? " = true" : " = false";
				break;
			case Enum:
			{
				Enumerated* pEnum = dynamic_cast<Enumerated*>(m_pType);
				if (pEnum != 0)	{
					std::vector<EnumValue>::const_iterator iter = pEnum->values.begin();
					while (iter != pEnum->values.end())	{
						if (m_Datum.m_int == (*iter).m_nValue)	{
							strFmt.Format(" = %s", (*iter).m_pName);
							break;
						}
						iter++;
					}
					break;
				}
			}
			// unknown enum value, use Char
			case Char:
			case UnsignedChar:
				strFmt.Format(" = '%c'", m_Datum.m_char);
				break;
			case Int:
				strFmt.Format(" = %d", m_Datum.m_int);
				break;
			case UnsignedInt:
				strFmt.Format(" = %u", m_Datum.m_unsigned);
				break;
			case Short:
				strFmt.Format(" = %hd", m_Datum.m_short);
				break;
			case UnsignedShort:
				strFmt.Format(" = %hu", m_Datum.m_unsignedshort);
				break;
			case Long:
				strFmt.Format(" = %ld", m_Datum.m_long);
				break;
			case UnsignedLong:
				strFmt.Format(" = %lu", m_Datum.m_unsignedlong);
				break;
			case Float:
				strFmt.Format(" = %g", m_Datum.m_float);
				break;
			case Double:
				strFmt.Format(" = %g", m_Datum.m_double);
				break;
			case LongDouble:
				strFmt.Format(" = %Lg", m_Datum.m_longdouble);
				break;
			default:
				break;
		}
	}
	strVarValue += strFmt;
}

void Item::ChangeMemoryVariable(HANDLE hProcess, const CString& strVarValue)
{
	if (m_nMemaddr == 0)
		return;		// not an lvalue

	TypeUnion tu;
	const char* nv = (LPCTSTR)strVarValue;

	if (isPointerOrArray())
		// ----- pointer type
		tu.T15 = (char*) atol(nv);
	else if (m_Type == Enum)	{
		if (isdigit(*nv))
			tu.T1 = atoi(nv);
		else	{
			Enumerated* pEnum = dynamic_cast<Enumerated*>(m_pType);
			if (pEnum != 0)	{
				std::vector<EnumValue>::const_iterator iter = pEnum->values.begin();
				while (iter != pEnum->values.end())	{
					if (strcmp(nv, (*iter).m_pName) == 0)	{
						tu.T1 = (*iter).m_nValue;
						break;
					}
					iter++;
				}
				if (iter == pEnum->values.end())	{
					// --- invalid entry
					AfxMessageBox("Invalid enum value", MB_ICONEXCLAMATION);
					return;
				}
			}
		}
	}
	else	{
		switch (m_StabType.typeno)	{
			case 1:
			case 4:
				// ---- int types
				tu.T1 = atoi(nv);
				break;
			case 2:
			case 10:
			case 11:
				// ---- char types
				tu.T2 = *nv;		
				break;
			case 3:
			case 5:
				// ---- long int types
				tu.T3 = atol(nv);	
				break;
			case 8:
			case 9:
				// ---- short int types
				tu.T8 = atoi(nv);
				break;
			case 12:
				// ---- float types
				tu.T12 = (float)atof(nv);
				break;
			case 13:
				tu.T13 = atof(nv);
				break;
			case 14:
				tu.T14 = atof(nv);
				break;
			default:
				break;
		}
	}

	int width = isPointerOrArray() ? sizeof(void*) : m_nWidth;
	WriteProcessMemory(hProcess, (void*)m_nMemaddr, &tu, width, 0);
}

void Item::BuildDisplay()
{
	int nInd = m_nIndirection;
	while (nInd--)
		m_strTypeDisp += "*";
	int nRef = m_nReferences;
	while (nRef--)
		m_strTypeDisp += "&";
	for (int i = 0; i < m_nDimensions; i++)	{
		CString strDim("[]");
		int nDim = *(m_pDimensions+i);
		if (nDim)
			strDim.Format("[%d]", nDim);
		m_strTypeDisp += strDim;
	}
}

Parser::Parser()
{
	m_pTokenBuffer = 0;
	m_pTokenPtr = 0;
	m_Token = 0;
	m_nConstant = 0;
}

Parser::~Parser()
{
	delete [] m_pTokenBuffer;
}

// ------- fetch the next token from the expression's source code
void Parser::NextToken()
{
    m_Token = *m_pTokenPtr++;
    m_strIdentifier.Empty();
    m_nConstant = 0;
    if (m_Token == '$') {
        m_nConstant = *(short int*)m_pTokenPtr;
        m_pTokenPtr += sizeof(short int);
    }
    else if (m_Token == '<')    {
        m_strIdentifier = m_pTokenPtr;
        m_pTokenPtr += strlen(m_pTokenPtr)+1;
    }
}

Item Parser::EvaluateExpression(const char *pExpression)
{
	Item item;
	delete [] m_pTokenBuffer;
	m_pTokenBuffer = new char[1024];
	m_pTokenPtr = m_pTokenBuffer;

	try	{
		LexicalScanner(pExpression);
		NextToken();
		item = AddSubtract();
		item.BuildDisplay();
	}
	catch (ParserError& pe)	{
		pe.DisplayMessage(pExpression);
	}
	return item;
}

Item Parser::AddSubtract()
{
	// Addition and subtraction
	Item item1 = MultiplyDivide();
	while (m_Token == '+' || m_Token == '-') 	{
        if (m_Token == '+')	{
			// -------- Addition: first operand
            NextToken();
			// ------ second operand
			Item item2 = MultiplyDivide();
			item1 += item2;
		}
        else	{
			// ---- Subtraction: first operand
            NextToken();
			// ------ second operand
			Item item2 = MultiplyDivide();
			item1 -= item2;
		}
	}
	return item1;
}

Item Parser::MultiplyDivide()
{
	Item item1 = Primary();

	while (m_Token == '*' || m_Token == '/') 	{
        if (m_Token == '*')	{
            NextToken();
			Item item2 = Primary();
			item1 *= item2;
		}
		else	{
            NextToken();
			Item item2 = Primary();
			item1 /= item2;
		}
	}
	return item1;
}

Item Parser::Primary()
{
	Item item1 = Element();
	while (m_Token == '[' || m_Token == '>' || m_Token == '.') 	{
		switch (m_Token)	{
			case '[':
				// ----- subscript operator
				do	{
					NextToken();
					Item item2 = AddSubtract();
					if (m_Token != ']')
						throw NeedRBracketError();
					item1 += item2;
					item1.DereferencePointer();
					NextToken();
				} while (m_Token == '[');
				break;
			case '>':
				// ----- pointer to member operator
				item1.DereferencePointer();
				// --- fall through
			case '.':
				// ----- structure member operator
				if (item1.GetItemType() != Item::Structure)
					throw NotStructureError();
				NextToken();
				if (m_Token != '<')
					throw MemberError();
				item1.GetMember(m_strIdentifier);
				NextToken();
				break;
			default:
				break;
		}
	}
	return item1;
}

Item Parser::Element()
{
	Item item;
	if (m_Token == '(')	{
		NextToken();
		item = AddSubtract();
		if (m_Token != ')')
			throw NeedRParenError();
		NextToken();
	}
	else	{
		switch (m_Token)	{
			case '*':
				// ---- pointer dereference operator
				NextToken();
				item = Primary();
				item.DereferencePointer();
				break;
			case '&':
				// ---- address of operator
				NextToken();
				item = Primary();
				item.AddressOf();
				break;
			case '$':
				// ---- integer constant
				item = Item(m_nConstant);
				NextToken();
				break;
			case '<':
				// ---- identifier
				item = Item(m_strIdentifier);
				NextToken();
				break;
			default:
				throw SyntaxError();
				break;
		}
	}
	return item;
}

// --- lexical scanner
void Parser::LexicalScanner(const char *pSrcbuf)
{
    char *tknptr = m_pTokenBuffer;

    while (*pSrcbuf)    {
        char c = *pSrcbuf++;    // next src code char
        c &= 0x7f;
        if (isspace(c))
            continue;
        switch (c)    {
            // --- operators
            case '>':
                --tknptr;
            case '+':
            case '-':
            case '*':
            case '/':
            case '[':
            case ']':
            case '(':
            case ')':
            case '.':
			case '&':
                *tknptr++ = c;
                break;
            default:
                if (isdigit(c))    {
                    // --- constant
                    --pSrcbuf;
                    Constant(&pSrcbuf, &tknptr);
                }
                else if (isalpha(c) || c == '_')    {
                    // --- identifier
                    *tknptr++ = '<';
                    --pSrcbuf;
                    while (isalpha(*pSrcbuf) || isdigit(*pSrcbuf) || *pSrcbuf == '_')
                        *tknptr++ = *pSrcbuf++;
                    *tknptr++ = 0;
                }
                else
                     // --- Bad character in input line
                    throw SyntaxError();
        }
    }
    *tknptr = '\0';
}

// --- Parse a decimal, octal or hexadecimal constant
void Parser::Constant(const char **srcstr, char **tknstr)
{
    const char *srcp;
    short int i;
    BOOL isDecimal = TRUE;

    srcp = *srcstr;
    if (*srcp == '0')    {
        if (isdigit(*(srcp+1)))    {
            // --- octal constant
            sscanf(srcp, "%ho", &i);
            while (isdigit(*srcp))
                ++srcp;
            isDecimal = FALSE;
        }
        else if (tolower(*(srcp+1)) == 'x')    {
            // --- hexadecimal constant
            srcp += 2;
            sscanf(srcp, "%hx", &i);
            while (isxdigit(*srcp))
                ++srcp;
            isDecimal = FALSE;
        }
    }
    if (isDecimal)    {
        // --- decimal integer number
        i = atoi(srcp);
        while (isdigit(*srcp))
            ++srcp;
    }
    *srcstr = srcp;

    **tknstr = '$'; // constant token
    ++(*tknstr);
    *((short int *)*tknstr) = i;
    *tknstr += sizeof(short int);
}

