// --------- QuincyDebug.cpp

#include "stdafx.h"
#include "Quincy.h"
#include "stabs.h"
#include "parser.h"
#include "textview.h"
#include "debugger.h"
#include "TraceDialog.h"

#define FLAG_TRACE_BIT 0x100

Debugger* Debugger::pDebugger;
BOOL Debugger::m_bSteppingTo;	// true if executing program is stepping to the cursor line
FileReference Debugger::steptobreakpoint;

Debugger::Debugger()
{
	ASSERT(pDebugger == 0);
	pDebugger = this;

	m_bProcessRunning = FALSE;
	m_bStepping = FALSE;
	m_bSteppingOver = FALSE;
	m_bSteppingOut = FALSE;
	m_bStopping = FALSE;
	m_bIsBroken = FALSE;
	m_bInProgram = FALSE;
	m_bTerminated = FALSE;
	m_bAccessViolation = FALSE;
	m_bFixBreakpoints = FALSE;

	m_hProcess = 0;
	m_hThread = 0;
	m_dwProcessId = 0;
	m_dwThreadId = 0;
	m_nCurrentLineNo = 0;

    m_nVoidType = 0;
    m_nBoolType = 0;

	m_hApplWnd = 0;
	m_nCurrentPC = 0;
	m_nBrokenPC = 0;
	m_nSteppingOutPC = 0;

	m_pCStabs = 0;
	m_DebugEvents[close]    = CreateEvent(0, FALSE,  FALSE, 0);
	m_DebugEvents[suspend]  = CreateEvent(0, FALSE, FALSE, 0);
	m_DebugEvents[resume]   = CreateEvent(0, FALSE, FALSE, 0);
	m_DebugEvents[active]   = CreateEvent(0, TRUE,  TRUE, 0);
	m_DebugEvents[breaking] = CreateEvent(0, FALSE, FALSE, 0);
	m_DebugEvents[finished] = CreateEvent(0, FALSE, FALSE, 0);
	m_nDrive = _getdrive();
	m_pOldPath = _getcwd(0, MAX_PATH);
}

Debugger::~Debugger()
{
	CloseHandle(m_DebugEvents[close]);
	CloseHandle(m_DebugEvents[suspend]);
	CloseHandle(m_DebugEvents[resume]);
	CloseHandle(m_DebugEvents[active]);
	CloseHandle(m_DebugEvents[breaking]);
	CloseHandle(m_DebugEvents[finished]);
	delete m_pCStabs;

	_chdrive(m_nDrive);
	_chdir(m_pOldPath);
	delete m_pOldPath;

	pDebugger = 0;
	m_bSteppingTo = FALSE;
}

void Debugger::CloseDebuggerHandle(HANDLE& hHandle)
{
	if (hHandle)	{
		CloseHandle(hHandle);
		hHandle = 0;
	}
}

void Debugger::ShutdownProcess()
{
	PostTerminationMessage();
	CloseDebuggerHandle(m_ProcessInformation.hThread);
	CloseDebuggerHandle(m_ProcessInformation.hProcess);

	CloseDebuggerHandle(m_hFile);

	theApp.CloseConsoleHandles();
	FreeConsole();
}

void Debugger::Execute(CString& strCmd)
{
	// --- determine where the debugged program should write its file output (if any)
	// --- if a project, see if the project has a working directory programmed
	CString strWk = theApp.GetProjectRuntimeDirectory();
	if (strWk.IsEmpty())	{
		// either not a project or no working directory programmed
		// see if the options have a working directory programmed
		strWk = theApp.GetRuntimeDirectory();
		if (strWk.IsEmpty())	{
			// no options working directory programmed
			if (theApp.IsOnCDROM(strCmd))
				// use the TEMP directory, if TEMP is set. (If not, you're on your own)
				strWk = theApp.GetTemporaryDirectory();
			else	{
				// ----- change to the subdirectory where 
				//       executable file is located
				int offset = strCmd.ReverseFind('\\');
				if (offset != -1)	{
					strWk = strCmd.Left(offset);
				}
			}
		}
	}
	if (!strWk.IsEmpty())
		_chdir(strWk);
	m_strCmd = strCmd;
	AfxBeginThread(&Debugger::DebugProg, 0);
}

void Debugger::ContinueDebugging()
{
	DWORD debugev = WaitForMultipleObjects(2, m_DebugEvents+4, FALSE, 0);
	if (debugev != WAIT_TIMEOUT)	{
		switch (debugev + 4 - WAIT_OBJECT_0)	{
			case breaking:
				theApp.BringToTop();
				SelectSteppingSourceLine();
				theApp.RefreshAllWatches();
				break;
			case finished:
				theApp.StopDebugger();
				// debugger is destroyed after this, don't do anything
				break;
			default:
				break;
		}
	}
}

UINT Debugger::DebugProg(void*)
{
	ASSERT(pDebugger != 0);
	pDebugger->DebugApplicationProgram();
	SetEvent(pDebugger->m_DebugEvents[finished]);
	return 0;
}

void Debugger::DebugApplicationProgram()
{
	InitStabs(m_strCmd);
	FixBreakpoints();

	ASSERT(theApp.m_pdlgTrace != 0);
	if (theApp.m_pdlgTrace->m_hWnd != 0)	{
		CListBox* pList = static_cast<CListBox*>(theApp.m_pdlgTrace->GetDlgItem(IDC_TRACELIST));
		if (pList != 0)	{
			pList->ResetContent();
			pList->Invalidate();
		}
	}
	CString strCmdLine;
	theApp.GetCommandLine(strCmdLine);

	BOOL bIsConsole = theApp.IsConsoleTarget();

	if (bIsConsole)	{
		AllocConsole();
		// ---- parse stdin redirection from the command line
		CString strRedirectedStdin;
		int index = strCmdLine.Find('<');
		if (index != -1)
			theApp.ParseFileSpec(strRedirectedStdin, strCmdLine, index+1);
		if (theApp.BuildStdin(strRedirectedStdin) == INVALID_HANDLE_VALUE)	{
			CString strErr("Cannot open redirected stdin ");
			strErr += strRedirectedStdin;
			AfxMessageBox(strErr, MB_ICONSTOP);
			return;
		}

		// ---- parse stdout redirection from the command line
		CString strRedirectedStdout;
		index = strCmdLine.Find('>');
		if (index != -1)
			theApp.ParseFileSpec(strRedirectedStdout, strCmdLine, index+1);
		if (theApp.BuildStdout(strRedirectedStdout) == INVALID_HANDLE_VALUE)	{
			CString strErr("Cannot create redirected stdout ");
			strErr += strRedirectedStdout;
			AfxMessageBox(strErr, MB_ICONSTOP);
			return;
		}
		theApp.BuildStderr();
	}

	m_strApplTitle = theApp.GetFileName(m_strCmd);

	CString strCmd(m_strCmd + " " + strCmdLine);

	DWORD dwflags = bIsConsole ? STARTF_USESTDHANDLES : 0;
	STARTUPINFO suinfo = {
		sizeof(STARTUPINFO),			// cb
		0, 0,							// reserved
		m_strApplTitle.GetBuffer(0),	// lpTitle
		0, 0,							// dwX, dwY
		0, 0,							// dwXSize, dwYSize
		0, 0,							// dwXCountChars, dwYCountChars
		0,								// dwFillAttribute
		dwflags,						// dwFlags
		0,								// wShowWindow
		0, 0,							// reserved
		theApp.m_hStdin,				// hStdInput;
		theApp.m_hStdout,				// hStdOutput;
		theApp.m_hStderr,				// hStdError;
	};

	DWORD dwCreationFlags = DEBUG_ONLY_THIS_PROCESS | DEBUG_PROCESS;

	m_bProcessRunning = 
		CreateProcess(	0,
						strCmd.GetBuffer(0),
						0,
						0,
						bIsConsole,
						dwCreationFlags,
						0,
						0,
						&suinfo,
						&m_ProcessInformation );

	if (m_bProcessRunning == FALSE)	{
		DWORD dwError = GetLastError();
		char msg[80+MAX_PATH*2];
		sprintf(msg, "Cannot execute program %s\n"
			         "Error code: %ld\n"
			         "Current path: ",
						strCmd.GetBuffer(0), dwError);
		_getcwd(msg+strlen(msg), MAX_PATH);
		if (bIsConsole)
			FreeConsole();
		AfxMessageBox(msg, MB_ICONSTOP);
	}
	else	{
		SetConsoleWindow();
		DebugProgram();
	}
}

void Debugger::DebugProgram()
{
	BOOL bDone = FALSE;

	while (!bDone)	{
		DWORD debugev = WaitForMultipleObjects(4, m_DebugEvents, FALSE, INFINITE);
		switch (debugev - WAIT_OBJECT_0)	{
			case close:
				bDone = TRUE;
				break;
			case suspend:
				// ---- stop the debuggee from running
				ResetEvent(m_DebugEvents[active]);
				// ---- notify the GUI that the debuggee has suspended
				SetEvent(m_DebugEvents[breaking]);
				break;
			case resume:
				SetEvent(m_DebugEvents[active]);
				SetThreadContext(m_hThread, &m_Context);
				ContinueDebugEvent(m_dwProcessId, m_dwThreadId, m_bTerminated ? DBG_EXCEPTION_NOT_HANDLED : DBG_CONTINUE);
				if ((m_Context.EFlags & FLAG_TRACE_BIT) == 0 && m_hApplWnd && !m_bTerminated)
					BringWindowToTop(m_hApplWnd);
				// --- fall through
			case active:
				RunDebugger();
				break;
		}
	}

	ShutdownProcess();
	CTextView* pView = theApp.GetTextView(m_strCurrentSrcFile);
	if (pView != 0)	{
		m_nCurrentLineNo = 0;
		pView->Invalidate();
		pView->UpdateWindow();
	}
}

BOOL CALLBACK Debugger::EnumWindowsProc(HWND hwnd, LPARAM pap)
{
	ASSERT(pDebugger != 0);
	CString* pApp = (CString*) pap;
	char title[MAX_PATH];
	::GetWindowText(hwnd, title, MAX_PATH);
	CString strTitle(title);
	if (strTitle == CString("QUINCY"))	{
		pDebugger->m_hApplWnd = hwnd;
		return FALSE;
	}
	return TRUE;
}

void Debugger::SetConsoleWindow()
{
	if (m_hApplWnd == 0)
		EnumWindows(EnumWindowsProc, (LPARAM) &m_strApplTitle);
}

void Debugger::RunDebugger()
{
	DEBUG_EVENT event;
	static BOOL bOnce = FALSE;
	m_bInProgram = TRUE;
	if (WaitForDebugEvent(&event, 0))	{
		switch (event.dwDebugEventCode)	{
			case CREATE_PROCESS_DEBUG_EVENT:
				m_dwProcessId = event.dwProcessId;
				m_dwThreadId = event.dwThreadId;
				m_hProcess = event.u.CreateProcessInfo.hProcess;
				m_hThread  = event.u.CreateProcessInfo.hThread;
				m_hFile  = event.u.CreateProcessInfo.hFile;

				// ---- build the runtime stepping breakpoint table
				// ---- and get all the breakpoint memory addresses
				LoadAllBreakpoints();
				bOnce = FALSE;
				break;

			case EXIT_PROCESS_DEBUG_EVENT:
				ShowPCCaret(FALSE);
				SetEvent(m_DebugEvents[close]);
				break;

			case EXCEPTION_DEBUG_EVENT:
				m_nCurrentPC =
					(unsigned long)(event.u.Exception.ExceptionRecord.ExceptionAddress);

				m_Context.ContextFlags = CONTEXT_FULL;
				GetThreadContext(m_hThread, &m_Context);

				switch (event.u.Exception.ExceptionRecord.ExceptionCode)	{
					case EXCEPTION_ACCESS_VIOLATION:
						m_bAccessViolation = TRUE;
						AfxMessageBox("Access violation", MB_ICONSTOP);
						if (m_pCStabs != 0)	{
							breakpoints.SetAllBreakpoints(m_hProcess, false);
							m_pCStabs->SourceFileAndLine(m_nCurrentPC, &m_strSteppingSrcFile, &m_nSteppingLineNo);
							m_bIsBroken = TRUE;
							m_bStepping = TRUE;
							m_nBrokenPC = m_nCurrentPC;
							m_bInProgram = FALSE;
							SetEvent(m_DebugEvents[suspend]);
						}
						else	{
							m_bTerminated = TRUE;
							SetEvent(m_DebugEvents[close]);
						}
						return;

					case EXCEPTION_SINGLE_STEP:
						if (m_nBrokenPC != 0 && m_nBrokenPC != m_nCurrentPC)	{
							breakpoints.SetBreakpoint(m_nBrokenPC, m_hProcess);
							m_nBrokenPC = 0;
						}
						break;

					case EXCEPTION_BREAKPOINT:
					{
						// --- there is one initial breakpoint to be ignored
						if (bOnce == FALSE)	{
							bOnce = TRUE;
							break;
						}

						if (m_bSteppingOut)	{
							m_bSteppingOut = FALSE;
							WriteProcessMemory(m_hProcess, (void*)m_nSteppingOutPC, &m_cSteppingOutInst, 1, 0);
							m_bStepping = TRUE;
							LoadAllBreakpoints();
						}
						else	{
							Breakpoint* pBrkpt = breakpoints.FindBreakpoint(m_nCurrentPC);
							if (pBrkpt != 0 && (pBrkpt->m_FileRef.m_nLineNo != m_nCurrentLineNo
									|| pBrkpt->m_FileRef.m_strFile != m_strSteppingSrcFile))	{
								// ---- turn off all breakpoints
								breakpoints.SetAllBreakpoints(m_hProcess, false);

								// ----- set the source file and line number where the break occurred
								m_strSteppingSrcFile = pBrkpt->m_FileRef.m_strFile;
								m_nSteppingLineNo = pBrkpt->m_FileRef.m_nLineNo;
								m_bIsBroken = TRUE;
								m_bInProgram = FALSE;

								// ---- suspend the debuggee
								SetEvent(m_DebugEvents[suspend]);
								return;
							}
							if (m_bTerminated)	{
								ShowPCCaret(FALSE);
								SetEvent(m_DebugEvents[close]);
								return;
							}
						}
						// ---- breaking at a line that was already broken with a different PC.
						// ---- or breaking at the end of a function to step out of the function
						// ---- clear the breakpoint's INT03 so we can execute the instruction
						breakpoints.SetBreakpoint(m_nCurrentPC, m_hProcess, false);
						// ---- adjust the instruction pointer back to the original op code
						--m_Context.Eip;
						SetThreadContext(m_hThread, &m_Context);
						break;
					}
					default:
						break;
				}
				break;

			default:
				break;
		}
		ContinueDebugEvent(event.dwProcessId, event.dwThreadId, DBG_CONTINUE);
	}
}

void Debugger::PostTerminationMessage()
{
	if (!m_bTerminated)	{
		m_bStopping = TRUE;
		if (m_hApplWnd != 0)	{
			BringWindowToTop(m_hApplWnd);
			SetForegroundWindow(m_hApplWnd);
		}
		char msg[] = "\nAny key to return to Quincy...";
		DWORD ct;
		TCHAR bf;
		WriteConsole(GetStdHandle(STD_ERROR_HANDLE), msg, sizeof msg - 1, &ct, 0);
		HANDLE hStdin = GetStdHandle(STD_INPUT_HANDLE);
		SetConsoleMode(hStdin, 0);
		FlushConsoleInputBuffer(hStdin);
		ReadConsole(hStdin, &bf, 1, &ct, 0);
		FlushConsoleInputBuffer(hStdin);
	}
}
// ------ resume a program that was interrupted by a breakpoint or step
void Debugger::ResumeProgram()
{
	if (m_bFixBreakpoints)
		FixBreakpoints();

	// ---- adjust the instruction pointer back to the breakpoint's original op code
	--m_Context.Eip;
	m_bIsBroken = FALSE;

	if (m_pCStabs != 0)	{
		if (!m_bSteppingOut)	{
			// --- reload the breakpoints
			LoadAllBreakpoints();
			// ---- clear the breakpoint's INT03 so we can execute the instruction
			breakpoints.SetBreakpoint(m_nCurrentPC, m_hProcess, false);
			if (!m_bSteppingTo)	{
				// ---- intercept next asm instruction so we can reset the breakpoint
				m_Context.EFlags |= FLAG_TRACE_BIT;
				m_nBrokenPC = m_nCurrentPC;
			}
			else	{
				m_bSteppingTo = FALSE;
				m_nBrokenPC = 0;
			}
		}
	}
	SetEvent(m_DebugEvents[resume]);
}
// ------- display the program counter in the source code when breaking or stepping
void Debugger::ShowPCCaret(BOOL bOnOff)
{
	if (m_nCurrentLineNo != 0)	{
		CTextView* pView = theApp.GetTextView(m_strCurrentSrcFile);
		if (pView != 0)
			pView->DrawMargins();
	}
}

void Debugger::SelectSteppingSourceLine()
{
	if (theApp.SelectFileAndLine(m_strSteppingSrcFile, m_nSteppingLineNo))	{
		ShowPCCaret(FALSE);
		m_strCurrentSrcFile = m_strSteppingSrcFile;
		m_nCurrentLineNo = m_nSteppingLineNo;
		ShowPCCaret(TRUE);
	}
}

void Debugger::FixBreakpoints()
{
	m_bFixBreakpoints = FALSE;
	std::set<FileReference>& fr = theApp.GetProgrammedBreakpoints();
	std::set<FileReference>::iterator iter;
	// --- vector of programmed breakpoints to erase following corrections
	std::vector<FileReference> dels;
	// --- vector of programmed breakpoints to insert following corrections
	std::vector<FileReference> inss;
	for (iter = fr.begin(); iter != fr.end(); iter++)	{
		// ----- if user programmed an invalid breakpoint (src file, lineno)
		//       try to find the next highest valid one
		CString fname = (*iter).m_strFile;
		int lno = (*iter).m_nLineNo;
		int highlno = m_pCStabs->HighestFileLineNumber(fname);
		MemoryAddress addr = 0;
		bool bfixing = false;
		bool bfixed = false;
		while (addr == 0 && lno <= highlno)	{
			addr = m_pCStabs->SourceLineAddress(fname, lno);
			if (addr != 0)
				bfixed = bfixing;
			else	{
				bfixing = true;
				lno++;
			}
		}
		if (addr == 0 && !bfixed)	{
			bfixing = false;
			while (addr == 0 && --lno > 0)	{
				addr = m_pCStabs->SourceLineAddress(fname, lno);
				if (addr != 0)
					bfixed = bfixing;
				else
					bfixing = true;
			}
		}
		if (addr == 0 || bfixed)	{
			if (m_bSteppingTo && fname.CompareNoCase(steptobreakpoint.m_strFile) == 0 && (*iter).m_nLineNo == steptobreakpoint.m_nLineNo)	{
				if (addr != 0)
					steptobreakpoint.m_nLineNo = lno;
				else
					// --- caret is on a non-executable source code line.
					//     just keep running.
					m_bSteppingTo = false;
				continue;
			}
			// ---- debugger is deleting this breakpoint, save old one's key to erase
			//      (can't do it now because of loop iteration)
			dels.push_back(*iter);
			if (addr != 0)
				// ---- save the new breakpoint to add
				inss.push_back(FileReference(fname, lno));
			CString strMsg;
			if (bfixed)	{
				strMsg.Format(	"Debugger is correcting breakpoint\n"
								"in %s\n"
								"from line %d to line %d",
							fname.GetBuffer(0), (*iter).m_nLineNo, lno);
				AfxMessageBox(strMsg, MB_ICONINFORMATION | MB_OK);
			}
			else	{
				strMsg.Format(	"Debugger is deleting breakpoint\n"
								"in %s\n"
								"at line %d",
							fname.GetBuffer(0), (*iter).m_nLineNo);
				AfxMessageBox(strMsg, MB_ICONINFORMATION | MB_OK);
			}
		}
	}
	// ---- delete the breakpoints that were on invalid lines
	std::vector<FileReference>::iterator del_iter;
	for (del_iter = dels.begin(); del_iter != dels.end(); del_iter++)	{
		iter = fr.find(*del_iter);
		if (iter != fr.end())	{
			fr.erase(iter);
			CTextView* pView = theApp.GetTextView((*del_iter).m_strFile);
			if (pView != 0)
				pView->SetBreakpointDisplay((*del_iter).m_nLineNo, FALSE);
		}
	}
	// ---- insert the breakpoints corrected to valid lines
	std::vector<FileReference>::iterator ins_iter;
	for (ins_iter = inss.begin(); ins_iter != inss.end(); ins_iter++)	{
		fr.insert(*ins_iter);
		CTextView* pView = theApp.GetTextView((*ins_iter).m_strFile);
		if (pView != 0)
			pView->SetBreakpointDisplay((*ins_iter).m_nLineNo, TRUE);
	}
}
// ---- build the runtime stepping breakpoint table
// ---- get all the breakpoint memory addresses
void Debugger::LoadAllBreakpoints()
{
	if (m_pCStabs == 0)
		return;
	breakpoints.ClearAllBreakpoints(m_hProcess);
	// ---- build the runtime stepping breakpoint table
	if (m_bStepping)	{
		const std::vector<LineAddr>& laddrtbl = m_pCStabs->GetLineAddressTable();
		std::vector<LineAddr>::const_iterator iter;

		MemoryAddress funcaddr = 0;
		if (m_bSteppingOver)	{
			// --- find the function address for the current broken line
			for (iter = laddrtbl.begin(); iter != laddrtbl.end(); iter++)	{
				if (m_strSteppingSrcFile.CompareNoCase((*iter).filename.c_str()) == 0)	{
					if (m_nSteppingLineNo == (*iter).lineno)	{
						funcaddr = (*iter).funcaddress;
						break;
					}
				}
			}
		}

		for (iter = laddrtbl.begin(); iter != laddrtbl.end(); iter++)	{
			if (m_bSteppingOver && funcaddr != (*iter).funcaddress)
				continue;
			FileReference fr((*iter).filename.c_str(), (*iter).lineno);
			breakpoints.LoadBreakpoint(fr, (*iter).funcaddress + (*iter).address);
		}
		m_bSteppingOver = FALSE;
	}
	else	{
		// ---- get the breakpoint memory addresses for the programmed breakpoints
		std::set<FileReference>& fr = theApp.GetProgrammedBreakpoints();
		bool bSteppingTo = false;
		if (m_bSteppingTo)	{
			FileReference stfr(steptobreakpoint.m_strFile, steptobreakpoint.m_nLineNo);
			if (fr.find(stfr) == fr.end())	{
				fr.insert(stfr);
				bSteppingTo = true;
			}
		}

		std::set<FileReference>::iterator iter;
		for (iter = fr.begin(); iter != fr.end(); iter++)	{
			CString fname = (*iter).m_strFile;
			int lno = (*iter).m_nLineNo;
			MemoryAddress addr = m_pCStabs->SourceLineAddress(fname, lno);
			ASSERT(addr != 0);
			breakpoints.LoadBreakpoint(FileReference(fname, lno), addr);
		}

		if (bSteppingTo)	{
			FileReference stfr(steptobreakpoint.m_strFile, steptobreakpoint.m_nLineNo);
			fr.erase(stfr);
		}
	}
#ifdef _DEBUG
	breakpoints.DumpBreakpoints();
#endif
	breakpoints.SetAllBreakpoints(m_hProcess);
}

void Debugger::InitStabs(const CString& strCmd)
{
    m_nVoidType = 0;
    m_nBoolType = 0;
	ASSERT(m_pCStabs == 0);
	m_pCStabs = new CStab(this);
	try {
		if (m_pCStabs->LoadStabs(strCmd))
			return;
	}
	catch (InvalidStab is)	{
		CString strMsg;
		strMsg.Format("Invalid debug info: Error %d (stab # %d)\n\"%s\"", 
			is.err, 
			m_pCStabs->StabLineNo(),
			(is.cp ? is.cp : "??")
		);
		AfxMessageBox(strMsg, MB_ICONEXCLAMATION);
	}
	// ---- no debugging information
	delete m_pCStabs;
	m_pCStabs = 0;
}

void Debugger::Stop()
{
	if (ProcessRunning() && !m_bStopping)	{
		m_bTerminated = TRUE;
		if (m_bAccessViolation || !isBroken())	{
			SetEvent(m_DebugEvents[close]);	// stopping a loop or getchar, no doubt
			ShowPCCaret(FALSE);
		}
		else	{
			StopStepping();
			ResumeProgram();
		}
	}
}

void Debugger::StepTo(const CString& strFile, int nLineNo)
{
	steptobreakpoint = FileReference(strFile, nLineNo);
	m_bSteppingTo = TRUE;
}

BOOL Debugger::ExecuteProgram()
{
	if (isBroken() && theApp.TestProgramChanged())
		return TRUE;
	if (isBroken())	{
		StopStepping();
		ResumeProgram();
	}
	else
		return FALSE;
	return TRUE;
}
// ------- the step command
BOOL Debugger::StepProgram()
{
	if (isBroken() && theApp.TestProgramChanged())
		return TRUE;
	if (isStepping())
		ResumeProgram();
	else	{
		SetStepping();
		if (isBroken())
			ResumeProgram();
		else
			return FALSE;
	}
	return TRUE;
}
// ----- the step over command
BOOL Debugger::StepOver() 
{
	if (isBroken())	{
		if (theApp.TestProgramChanged())
			return TRUE;
		m_bSteppingOver = TRUE;
		SetStepping();
		ResumeProgram();
		return TRUE;
	}
	return FALSE;
}

// ----- the step out command
BOOL Debugger::StepOut() 
{
	if (isBroken())	{
		if (theApp.TestProgramChanged())
			return TRUE;

		m_nSteppingOutPC = m_pCStabs->FunctionExitAddress(m_nCurrentPC);
		if (m_nSteppingOutPC != 0)	{
			m_bSteppingOut = TRUE;
			ReadProcessMemory(m_hProcess, (void*)m_nSteppingOutPC, &m_cSteppingOutInst, 1, 0);
			WriteProcessMemory(m_hProcess, (void*)m_nSteppingOutPC, "\xcc", 1, 0);
			StopStepping();
		}
		else
			SetStepping();

		ResumeProgram();
		return TRUE;
	}
	return FALSE;
}
