// ------- stabs.cpp

#include "stdafx.h"
#include <typeinfo>
#include "Quincy.h"
#include "debugger.h"
#include "stabs.h"


class foobar {};


TypeIndex::TypeIndex(int fno, int hno, int tno) : fileno(fno), hdrno(hno), typeno(tno)
{
	if (fileno != 0 && hdrno == 0 && typeno < 19)
		fileno = 1;
}
bool TypeIndex::operator==(const TypeIndex& x) const
{
	return fileno == x.fileno && hdrno == x.hdrno && typeno == x.typeno;
}
bool TypeIndex::operator<(const TypeIndex& x) const
{
	if (fileno == x.fileno)
		if (hdrno == x.hdrno)
			return typeno < x.typeno;
		else
			return hdrno < x.hdrno;
	return fileno < x.fileno;
}
// ----- table of intrinsic data type descriptions
//       these are used for watch and examine variables
//       don't use sizeof because these are GNU widths, not VC++ widths
int CStab::nIntrinsicTypeWidths[] = {
	4,	//  1. int
	1,  //  2. char
	4,	//  3. long int
	4,	//  4. unsigned int
	4,	//  5. long unsigned int
	8,	//  6. long long int          (nonstandard GCC type)
	8,	//  7. long long unsigned int (nonstandard GCC type)
	2,	//  8. short int
	2,	//  9. short unsigned int
	1,	// 10. signed char
	1,	// 11. unsigned char
	4,	// 12. float
	8,	// 13. double
	8,	// 14. long double
};

CStab::CStab(Debugger* pDebugger)
{
	m_pDebugger = pDebugger;
	InitializeTables();
}

CStab::~CStab()
{
	DeleteTables();
}

void CStab::DeleteTables()
{
	delete [] m_pStabStrings;
	delete [] m_pStabs;

	std::map<TypeIndex, Type*>::reverse_iterator iter = m_Types.rbegin();
	while (iter != m_Types.rend())
		delete (*iter++).second;
}

void CStab::InitializeTables()
{
	m_pStabStrings = 0;
	m_pStabs = 0;

	m_nStabno = 0;
	m_nNbrStabs = 0;
	m_nFileno = 0;
}

// ----- increment the pointer to a stab string, stepping over line continuation characters
void CStab::IncrStrPtr(char** cp, int n)
{
	while (n--)	{
		(*cp)++;
		if (**cp == '\\')
			*cp += 2;
	}
}
inline void CStab::BypassSpaces(char** cp)
{
	while (**cp == ' ' || **cp == '\t')
		IncrStrPtr(cp);
}

inline void CStab::BypassDigits(char** cp)
{
	while (isdigit(**cp))
		IncrStrPtr(cp);
}

inline void CStab::BypassNonDigits(char** cp)
{
	while (!isdigit(**cp))
		IncrStrPtr(cp);
}

inline BOOL CStab::isIdentifierCharacter(char ch)
{
	return (isdigit(ch) || isalpha(ch) || ch == '_');
}

void CStab::BypassIdentifier(char** cp)
{
	if (**cp == ':' && *((*cp)+1) == ':')
		IncrStrPtr(cp, 2);

	if (**cp == '$')
		IncrStrPtr(cp);

	if (**cp != ' ')	// stab bug in .c files for unnamed enums
		if (!isIdentifierCharacter(**cp))
			throw InvalidStab(__LINE__, *cp);

	while (isIdentifierCharacter(**cp) || **cp == '$' || **cp == ' ')
		IncrStrPtr(cp);
	if (**cp == '<')	{
		// ------- template arguments
		do	{
			IncrStrPtr(cp);		// bypass '<' first time, ',' other times
			BypassSpaces(cp);
			if (**cp == '>')
				break;
			BypassIdentifier(cp);
			BypassSpaces(cp);
			while (**cp == '&' || **cp == '*')	{
				IncrStrPtr(cp);
				BypassSpaces(cp);
				if (strncmp(*cp, "const", 5) == 0)	{
					(*cp) += 5;
					BypassSpaces(cp);
				}
			}
			BypassSpaces(cp);
		} while (**cp == ',');
		if (**cp == ' ')
			IncrStrPtr(cp);
		NeedCharacter(cp, '>', __LINE__);
		IncrStrPtr(cp);
	}
	if (**cp == ':' && *((*cp)+1) == ':' && isIdentifierCharacter(*((*cp)+2)))
		BypassIdentifier(cp);
}

void CStab::BypassTypeIndex(char** cp)
{
	IncrStrPtr(cp);			// bypass '('
	BypassDigits(cp);		// bypass file number
	IncrStrPtr(cp);			// bypass ','
	BypassDigits(cp);		// bypass type number
	IncrStrPtr(cp);			// bypass ')'
}

// ----- extracts the index of a generic Type
TypeIndex CStab::ExtractGenericType(char** cp)
{
	NeedCharacter(cp, '(', __LINE__);
	IncrStrPtr(cp);			// bypass '('
	int nHdr = atoi(*cp);	// header file number
	BypassDigits(cp);
	IncrStrPtr(cp);			// bypass ','
	int nType = atoi(*cp);	// type number
	BypassDigits(cp);
	IncrStrPtr(cp);			// bypass ')'
	TypeIndex type(m_nFileno, nHdr, nType);
	return type;
}

void CStab::SkipToEndOfMembers(char** cp)
{
	while (**cp)	{
		if (**cp == ';')	{
			IncrStrPtr(cp);
				if (**cp == ';')	{
					IncrStrPtr(cp);
					break;
				}
			continue;
		}
		IncrStrPtr(cp);
	}
}
TypeIndex  CStab::AccumulateStructType(char** cp, TypeIndex typex) throw (InvalidStab)
{
	// defining a struct type
	IncrStrPtr(cp);		// bypass 's' or 'u'
	Struct* pStruct = 0;

	if (m_Types.find(typex) == m_Types.end())	{
		// ------- unnamed struct or union
		pStruct = new Struct;
	}
	else	{
		Intrinsic* pInt = dynamic_cast<Intrinsic*>(m_Types[typex]);
		if (pInt == 0)	{
			// --- struct was already defined as something else (??)
			SkipToEndOfMembers(cp);
			return typex;
		}

		pStruct = new Struct(pInt->m_pName);
		delete pInt;	// discard the Intrinsic object prematurely built for this struct
	}

	m_Types[typex] = pStruct;
	// ------- struct width
	pStruct->m_nWidth = atoi(*cp);
	BypassDigits(cp);	// bypass width


	if (**cp == '!')	{
		// --- base class info
		IncrStrPtr(cp);		// bypass '!'
		int nBases = atoi(*cp);	// # of base classes
		BypassDigits(cp);
		IncrStrPtr(cp);		// bypass ','

		for (int i = 0; i < nBases; i++)	{
			IncrStrPtr(cp);			// bypass '0' or '1' virtual character
			IncrStrPtr(cp);			// bypass '0', '1' or '2' visibility character
			BypassDigits(cp);		// bypass offset
			IncrStrPtr(cp);			// bypass ','
			AccumulateTypes(0, cp);	// bypass base class type descriptor
			IncrStrPtr(cp);			// bypass ';'
		}
	}

	do	{
		char* nm = *cp;	// member name

		if (strncmp(*cp, "$v", 2) == 0)	{
			// ---- vtbl entry
			IncrStrPtr(cp, 3);			// bypass "$vf" or "$vb"
			BypassTypeIndex(cp);		// vtbl type id
			IncrStrPtr(cp);				// bypass ':'
			AccumulateTypes(nm, cp);
			IncrStrPtr(cp);		// bypass ','
			BypassDigits(cp);	// bypass bit offset
			IncrStrPtr(cp);		// bypass ';'
			continue;
		}

		// ---- bypass member name
		if (**cp == ':' && *((*cp)+1) != ':')
			throw foobar();		// apparent bug in winnt.h, etc. stabs generation

		BypassIdentifier(cp);

		**cp = '\0';			// null-terminate the name
		IncrStrPtr(cp);			// bypass what was ':' and is now '\0'

		if (**cp == ':')
			// ------ member function
			IncrStrPtr(cp);		// bypass ':'

		if (**cp == '/')
			IncrStrPtr(cp, 2);	// bypass access specifier

		// ---- get the type of the member
		TypeIndex ndx = AccumulateTypes(nm, cp);

		if (**cp == ':')	{
			char* cpp = *cp;
			if (*(cpp+1) == '_')	{
				// --- static data member
				IncrStrPtr(cp);					// bypass ':'
				while (**cp != ';')
					IncrStrPtr(cp);				// bypass static mangled name
				continue;
			}
			while (**cp && **cp != ';')	{
				// ----- member function
				IncrStrPtr(cp);					// bypass ':'
				while (**cp && **cp != ';')
					IncrStrPtr(cp);				// bypass argument characters
				NeedCharacter(cp, ';', __LINE__);
				IncrStrPtr(cp);					// bypass ';'
				if (**cp == '$')
					break;						// this is a static data member
				if (!isdigit(**cp))
					continue;					// static member function
				IncrStrPtr(cp, 2);				// bypass '2A'
				if (**cp == '*')	{
					// --- virtual member function
					IncrStrPtr(cp);				// bypass '*'
					if (**cp == '-')
						IncrStrPtr(cp);			// bypass '-'
					BypassDigits(cp);			// bypass vtbl offset
					IncrStrPtr(cp);				// bypass ';'
					BypassTypeIndex(cp);		// bypass base class index
					if (**cp == '=')	{
						IncrStrPtr(cp);			// bypass '='
						BypassIdentifier(cp);
						IncrStrPtr(cp);			// bypass ':'
					}
					IncrStrPtr(cp);				// bypass ';'

				}
				else
					IncrStrPtr(cp);					// bypass '.'
				if (**cp != ';')	{
					NeedCharacter(cp, '(', __LINE__);
					AccumulateTypes(0, cp);
					NeedCharacter(cp, ':', __LINE__);
				}
			}
			if (**cp != '$')	{
				NeedCharacter(cp, ';', __LINE__);
				IncrStrPtr(cp);					// bypass ';'
			}
		}
		else	{
			// ------ data member
			if (**cp == ';')
				// --- the member was an unnamed struct or union
				IncrStrPtr(cp);					// bypass ';'

			IncrStrPtr(cp);		// bypass ','
			// --- member's struct offset
			int offset = atoi(*cp) / 8;

			BypassDigits(cp);	// bypass bit offset
			IncrStrPtr(cp);		// bypass ','
			BypassDigits(cp);	// bypass bit width
			IncrStrPtr(cp);		// bypass ';'
			Member mbr(nm, ndx, offset);
			pStruct->members.push_back(mbr);
		}
	} while (**cp != ';');
	IncrStrPtr(cp);		// bypass ';'
	if (strncmp(*cp, "~%", 2) == 0)	{
		// --- class contains virtual functions
		IncrStrPtr(cp,2);			// bypass '~%'
		BypassTypeIndex(cp);		// bypass first base class index
		IncrStrPtr(cp);		// bypass ';'
	}
	return typex;
}

TypeIndex CStab::AccumulateMemberFunction(char** cp, TypeIndex typex) throw (InvalidStab)
{
	IncrStrPtr(cp); // bypass '#'
	if (**cp == '#')	{
		IncrStrPtr(cp); // bypass 2nd '#'
		TypeIndex ndx = AccumulateTypes(0, cp);
		NeedCharacter(cp, ';', __LINE__);
		IncrStrPtr(cp); // bypass ';'
		return ndx;
	}
	NeedCharacter(cp, '(', __LINE__);
	BypassTypeIndex(cp);	// bypass class type
	NeedCharacter(cp, ',', __LINE__);
	IncrStrPtr(cp);			// bypass ','
	NeedCharacter(cp, '(', __LINE__);
	AccumulateTypes(0, cp);	// return type
	NeedCharacter(cp, ',', __LINE__);
	IncrStrPtr(cp);			// bypass ','
	while (**cp && **cp != ';')	{
		NeedCharacter(cp, '(', __LINE__);
		AccumulateTypes(0, cp);	// parameter type
		if (**cp == ',')
			IncrStrPtr(cp);
	}
	NeedCharacter(cp, ';', __LINE__);
	IncrStrPtr(cp);			// bypass ';'
	return typex;
}

TypeIndex CStab::AccumulateEnumType(char** cp, TypeIndex typex) throw(InvalidStab)
{
	IncrStrPtr(cp); // bypass 'e'

	if (m_Types.find(typex) == m_Types.end())
		throw InvalidStab(__LINE__, *cp);

	Intrinsic* pInt = dynamic_cast<Intrinsic*>(m_Types[typex]);
	Enumerated* pEnum = 0;
	if (pInt != 0)	{
		pEnum = new Enumerated(pInt->m_pName);
		delete pInt;	// discard the Intrinsic object prematurely built for this enum
		m_Types[typex] = pEnum;
	}
	do	{
		char* nm = *cp;
		// ---- bypass enum value name
		while (isIdentifierCharacter(**cp))
			IncrStrPtr(cp);
		**cp = '\0';			// null-terminate the name
		IncrStrPtr(cp);			// bypass ':'
		int val = atoi(*cp);
		if (**cp == '-')
			IncrStrPtr(cp);		// bypass '-'
		BypassDigits(cp);		// bypass the value
		IncrStrPtr(cp);			// bypass ','
		if (pEnum != 0)	{
			EnumValue ev(nm,val);
			pEnum->values.push_back(ev);
		}
	} while (**cp != ';');
	return typex;
}

TypeIndex CStab::AccumulateRangeType(char** cp)
{
	// range type
	IncrStrPtr(cp);		// bypass 'r'
	TypeIndex ndx = ExtractGenericType(cp);
	IncrStrPtr(cp);		// bypass ';'
	if (**cp == '-')
		IncrStrPtr(cp);	// bypass '-'
	BypassDigits(cp);	// bypass 1st range value
	IncrStrPtr(cp);		// bypass ';'
	if (**cp == '-')
		IncrStrPtr(cp);	// bypass '-'
	BypassDigits(cp);	// bypass 2nd range value
	IncrStrPtr(cp);		// bypass ';'
	return ndx;
}

TypeIndex CStab::AccumulatePointerType(char** cp, TypeIndex typex) throw (InvalidStab)
{
	// pointer type
	IncrStrPtr(cp);
	TypeIndex ndx = ExtractGenericType(cp);
	Type* pType = 0;

	if (m_Types.find(typex) != m_Types.end())	{
		// --- a typedef
		pType = m_Types[typex];
		Type* pt = pType;
		Intrinsic* pInt = dynamic_cast<Intrinsic*>(pType);
		if (pInt != 0)
			pType = new Typedef(pInt->m_pName, ndx);
		else
			pType =  new Pointer(ndx);
		delete pt;
	}
	else
		pType =  new Pointer(ndx);
	m_Types[typex] = pType;
	return ndx;
}

TypeIndex CStab::AccumulateReferenceType(char** cp, TypeIndex typex) throw (InvalidStab)
{
	// reference type
	IncrStrPtr(cp);
	TypeIndex ndx = ExtractGenericType(cp);
	Type* pType = 0;

	if (m_Types.find(typex) != m_Types.end())	{
		// --- a typedef
		pType = m_Types[typex];
		Type* pt = pType;
		Intrinsic* pInt = dynamic_cast<Intrinsic*>(pType);
		if (pInt != 0)
			pType = new Typedef(pInt->m_pName, ndx);
		else
			pType =  new Reference(ndx);
		delete pt;
	}
	else
		pType =  new Reference(ndx);
	m_Types[typex] = pType;
	return ndx;
}

TypeIndex  CStab::AccumulateParameterType(char* pName, char** cp)
{
	IncrStrPtr(cp);
	TypeIndex typex = ExtractGenericType(cp);
	return typex;
}

TypeIndex CStab::AccumulateArrayType(char** cp, TypeIndex typex) throw(InvalidStab)
{
	// defining an array type

	if (typex.typeno == 0)
		throw InvalidStab(__LINE__, *cp);

	// ---- bypass unneeded "ar(n,n);0;"
	IncrStrPtr(cp, 3);	// bypass "ar("
	BypassDigits(cp);	// bypass n
	IncrStrPtr(cp);		// bypass ','
	BypassDigits(cp);	// bypass n
	IncrStrPtr(cp,2);	// bypass ');'
	BypassDigits(cp);	// bypass n
	IncrStrPtr(cp);		// bypass ';'
	
	int nDimension = atoi(*cp) + 1;

	if (nDimension == 0)
		IncrStrPtr(cp);		// bypass '-'

	BypassDigits(cp);	// bypass dimension
	IncrStrPtr(cp);		// bypass ';'

	TypeIndex ndx = ExtractGenericType(cp);
	if (m_Types.find(typex) != m_Types.end())	{
		Type* pType = m_Types[typex];
		Intrinsic* pInt = dynamic_cast<Intrinsic*>(pType);
		if (pInt == 0)	{
			Array* pArr = dynamic_cast<Array*>(pType);
			if (pArr == 0)	{
				// --- could be converting previous function parameter pointer to array. ugh.
				Pointer* pPtr = dynamic_cast<Pointer*>(pType);
				if (pPtr == 0)
					throw InvalidStab(__LINE__, *cp);
			}
		}
		delete pType;
	}
	Array* pArr = new Array(ndx, nDimension);
	m_Types[typex] = pArr;
	return ndx;
}

TypeIndex CStab::AccumulateIntrinsicType(char* pName, char** cp)
{
	IncrStrPtr(cp);
	TypeIndex typex = ExtractGenericType(cp);
	if (m_Types.find(typex) == m_Types.end())	{
		int len = 0;	// could be struct, void, enum, unsupported gnu type, etc.
		if (typex.hdrno == 0 && typex.typeno <= sizeof(nIntrinsicTypeWidths) / sizeof(int))
			len = nIntrinsicTypeWidths[typex.typeno - 1];	
		Intrinsic* pInt = new Intrinsic(pName, len); 
		m_Types[typex] = pInt;
	}
	return typex;
}

TypeIndex CStab::AccumulateType(char* pName, char** cp, TypeIndex typex) throw(InvalidStab)
{
	switch (**cp)	{
		case '@':
			// ---- probably the nonstandard "@s8:-16;" thingie for bool
			if (strcmp(pName, "bool") == 0)	{
				if (m_Types.find(typex) != m_Types.end())	{
					Intrinsic* pInt = dynamic_cast<Intrinsic*>(m_Types[typex]);
					if (pInt != 0)
						pInt->m_nWidth = 1;
				}
			}
			break;
		case '&':
			// ---- reference type
			typex = AccumulateReferenceType(cp, typex);
			break;
		case '*':
			// ---- pointer type
			typex = AccumulatePointerType(cp, typex);
			break;
		case 'p':
			// ---- parameter type
			if (pName == 0)
				throw InvalidStab(__LINE__, *cp);
			typex = AccumulateParameterType(pName, cp);
			break;
		case 'r':
			// ---- range type
			typex = AccumulateRangeType(cp);
			break;
		case 'a':
			// ---- array type
			typex = AccumulateArrayType(cp, typex);
			break;
		case 'e':
			// ---- enum type
			typex = AccumulateEnumType(cp, typex);
			break;
		case 'x':
			// ---- forward reference
			IncrStrPtr(cp, 2);		// bypass "xs" or "xu"
			BypassIdentifier(cp);
			**cp = '\0';
			IncrStrPtr(cp);			// bypass ':'
			break;
		case 's':
		case 'u':
			// ---- class/structure/union definition
			typex = AccumulateStructType(cp, typex);
			break;
		case 'T':
			// ---- user-defined type
			if (pName == 0)
				throw InvalidStab(__LINE__, *cp);
			if (*((*cp)+1) == 't')
				// Tt(n.n) ???
				IncrStrPtr(cp);
			typex = AccumulateIntrinsicType(pName, cp);
			break;
		case 't':
			// ---- intrinsic type
			if (pName == 0)
				throw InvalidStab(__LINE__, *cp);
			typex = AccumulateIntrinsicType(pName, cp);
			break;
		case 'f':	// static function (including static member functions) (what about function pointers?)
		case 'F':	// global function
		case 'S':	// static global variable
		case 'V':	// static local variable
		case 'G':	// global variable
			IncrStrPtr(cp);
			typex = ExtractGenericType(cp);
			break;
		case '(':
		{
			TypeIndex ndx = ExtractGenericType(cp);
			if (m_Types.find(typex) != m_Types.end() && ndx != typex)	{
				Type* pType = m_Types[typex];
				delete pType;
				Typedef* pTyp = new Typedef(pName, ndx);
				m_Types[typex] = pTyp;
			}
			typex = ndx;
			break;
		}
		case '#':
			// ---- member function
			typex = AccumulateMemberFunction(cp, typex);
			break;
		default:
			throw InvalidStab(__LINE__, *cp);
			break;
	}
	return typex;
}

TypeIndex CStab::AccumulateTypes(char* pName, char** cp)
{
	TypeIndex typex = AccumulateType(pName, cp);
	if (**cp == '=')	{
		TypeIndex ndx = typex;
		do	{
			IncrStrPtr(cp);
			ndx = AccumulateType(pName, cp, ndx);
		} while (**cp == '=');
	}
	return typex;
}

// ----- initialize and load stabs for an executable program
BOOL CStab::LoadStabs(const CString& strCmd) throw (InvalidStab)
{
	// ---- initialize the stab object
	DeleteTables();
	InitializeTables();

	// ---- open the executable file
    std::ifstream ifile(strCmd, std::ios::binary);
	if (ifile.fail())
		return FALSE;	// no such exe file

	// ----- read the DOS header
	IMAGE_DOS_HEADER dosheader;
	ifile.read((char*)&dosheader, sizeof dosheader);

	// ----- test .exe file signature
	if (dosheader.e_magic != IMAGE_DOS_SIGNATURE)
		return FALSE;	// not an exe file

	// ----- read offset to portable executable header
	ifile.seekg(dosheader.e_lfanew);
	// ---- read the pertable executable header
	struct {
		long signature;
		IMAGE_FILE_HEADER fhdr;
		IMAGE_OPTIONAL_HEADER ohdr;
	} header;

	ifile.read((char*)&header, sizeof header);
	// ----  check the signature
	if (header.signature != IMAGE_NT_SIGNATURE)
		return FALSE;	// not a portable exe file

	long nBaseAddress = header.ohdr.ImageBase;	// program's base address
	long nDataVirtAddr = 0;						// .data section's virtual address
	int  nDataSectionNo = 0;					// .data section's number

	// ----- iterate through the segments looking for .stab and .stabstr
	unsigned long staboffset = 0, stabstroffset = 0, stabstrlen = 0, stabslen = 0;
	for (int i = 0; i < header.fhdr.NumberOfSections; i++)	{
		IMAGE_SECTION_HEADER secthdr;
		ifile.read((char*)&secthdr, sizeof(secthdr));
		char name[9];
		strncpy(name, (char*)secthdr.Name, 8);
		name[8] = '\0';


		if (strcmp(name, ".stab") == 0)	{
			// --- found .stab, get length and offset to stab table
			stabslen = secthdr.Misc.VirtualSize;
			staboffset = secthdr.PointerToRawData;
		}
		else if (strncmp(name, ".stabstr", 8) == 0)	{
			// --- found .stabstr, get length and offset to stab strings
			stabstrlen = secthdr.Misc.VirtualSize;
			stabstroffset = secthdr.PointerToRawData;
		}
		else if (strncmp(name, ".data", 8) == 0)	{
			nDataVirtAddr = secthdr.VirtualAddress;
			nDataSectionNo = i + 1;
		}
	}
	if (staboffset == 0 || stabstroffset == 0 || stabslen == 0 || stabstrlen == 0)
		return FALSE;	// incomplete or no symbol information in this portable exe file

	// ------ read the stabs
	m_nNbrStabs = stabslen / sizeof(StabEntry);
	m_pStabs = new StabEntry[m_nNbrStabs];
	ifile.seekg(staboffset);
	ifile.read((char*)m_pStabs, stabslen);

	// ----- read the strings
	m_pStabStrings = new char[stabstrlen];
	ifile.seekg(stabstroffset);
	ifile.read((char*)m_pStabStrings, stabstrlen);

	// ----- load the symbol table's string table
	long symstroffset = header.fhdr.PointerToSymbolTable + 
						header.fhdr.NumberOfSymbols * sizeof(IMAGE_SYMBOL);
	ifile.seekg(symstroffset);
	long symstrlen;
	// --- get the length of the symbol string table (which includes the length field itself?)
	ifile.read((char*)&symstrlen, sizeof symstrlen);
	symstrlen -= sizeof symstrlen;
	char* symstrings = new char[symstrlen];
	ifile.read(symstrings, symstrlen);

	// ----- load the pertinant symbols into the symbol table
	IMAGE_SYMBOL isym;

	ifile.seekg(header.fhdr.PointerToSymbolTable);
	int symbypass = 0;
	for (int is = 0; is < header.fhdr.NumberOfSymbols; is++)	{
		ifile.read((char*)&isym, sizeof(IMAGE_SYMBOL));
		if (symbypass)
			--symbypass;
		else	{
			symbypass = isym.NumberOfAuxSymbols;
			if (isym.Type == IMAGE_SYM_TYPE_NULL && 
					isym.SectionNumber == nDataSectionNo &&
						isym.StorageClass == IMAGE_SYM_CLASS_EXTERNAL)	{
				char name[9];
				char* np;
				if (isym.N.Name.Short != 0)	{
					strncpy(name, (char*)isym.N.ShortName, 8);
					name[8] = '\0';
					np = name;
				}
				else
					np = symstrings + isym.N.Name.Long - sizeof(long);
				if (*np == '_')
					np++;
				std::string sname(np);
				MemoryAddress addr = nBaseAddress + nDataVirtAddr + isym.Value;
				imagesymbols[sname] = addr;
			}
		}
	}

	delete [] symstrings;

	// --------- collect stab data
	m_nCurrentFileScope = 0;
	m_nGlobalScopeEntry = 0;

	int highmem = IterateStabs(0, m_nNbrStabs);

	if (scopestack.size() != 0)
		throw InvalidStab(__LINE__, 0);
	PostExitScope(highmem, Symbol::file);
	PostExitScope(highmem, Symbol::global);

	// ----- finished with the raw stabs data
	delete [] m_pStabs;
	m_pStabs = 0;

#ifdef _DEBUG
//	DumpTypes();
//	DumpSourceLine();
//	DumpSymbols();
#endif

	return TRUE;
}

int CStab::IterateStabs(int nBegin, int nEnd)
{
	std::string currentpath;
	std::stack<std::string> filenames;		// current file name
	int nCurrentFunctionScope = 0;

	int nFunctionAddress = 0;

	int highmem = 0;
	int lineaddress = 0;
	// ----- iterate stabs for the program
	for (m_nStabno = nBegin; m_nStabno < nEnd; m_nStabno++)	{

		try	{
			// ---- if this is a continuation stab, it's already been processed
			if (isContinuation(m_pStabs[m_nStabno]))
				continue;

			// --- if this iteration was launched from this stab in a previous iteration,
			//        skip this stab
			if (skippers.find(m_nStabno) != skippers.end())
				continue;

			switch (m_pStabs[m_nStabno].m_nType)	{
				case N_SOL:	// included source file
				{
					if (filenames.size() == 0)
						throw InvalidStab(__LINE__, 0);

					nFunctionAddress = m_pStabs[m_nStabno].m_nValue;

					unsigned long int noffset = m_pStabs[m_nStabno].m_nStringOffset;
					filenames.pop();
					std::string file;
					char* cp = m_pStabStrings + noffset;
					if (*cp != '/' || *(cp+1) != '/')
						file = currentpath;
					file += std::string(cp);
					FixPath(file);
					filenames.push(file);
					break;
				}
				case N_SO:	// source file
				{

					nFunctionAddress = m_pStabs[m_nStabno].m_nValue;

					unsigned long int noffset = m_pStabs[m_nStabno].m_nStringOffset;
					char* cp = m_pStabStrings + noffset;
					if (*cp)	{
						if (*(cp + strlen(cp)-1)  == '/')
							// --- source file path stab
							currentpath = cp;
						else	{
							// ----- N_SO for source file name; save string offset for 
							//       subsequent source file, line number table entries
							std::string file;
							if (*cp != '/' || *(cp+1) != '/')
								file = currentpath;
							file += std::string(cp);
							FixPath(file);
							filenames.push(file);
							// ---- store current file scope
							m_nCurrentFileScope = m_pStabs[m_nStabno].m_nValue;
							// ---- store global scope
							if (m_nGlobalScopeEntry == 0)
								m_nGlobalScopeEntry = m_nCurrentFileScope;
							// ---- post exit file scope to file scope symbols for this file
							PostExitScope(m_nCurrentFileScope, Symbol::file);
							m_nFileno++;
						}
					}
					else	{
						// ---- end of C/CPP source file
						filenames.pop();
						if (!filenames.empty())
							throw InvalidStab(__LINE__, 0);
					}
					break;
				}
				case N_EXCL:	// deleted duplicate include file to optimize exe size
				{
					// --- search for matching N_BINCL
					int val = m_pStabs[m_nStabno].m_nValue;
					int i, j;
					for (i = 0; i < m_nStabno; i++)	{
						if (m_pStabs[i].m_nType == N_BINCL && m_pStabs[i].m_nValue == val)	{
							if (strcmp(m_pStabStrings + m_pStabs[i].m_nStringOffset,
									m_pStabStrings + m_pStabs[m_nStabno].m_nStringOffset) == 0)	{
								// --- find end of included file
								int incct = 0;
								for (j = i; j < nEnd; j++)	{
									if (m_pStabs[j].m_nType == N_BINCL)
										incct++;
									else if (m_pStabs[j].m_nType == N_EINCL)
										if (--incct == 0)
											break;
								}
								if (j == nEnd)
									throw InvalidStab(__LINE__, 0);
								int savstabno = m_nStabno;
								// --- skip this stab, subsequent iterations
								skippers.insert(m_nStabno);
								IterateStabs(i, j + 1);
								m_nStabno = savstabno;
								break;
							}
						}
					}
					if (i == m_nStabno)
						throw InvalidStab(__LINE__, 0);
					break;
				}
				case N_BINCL:	// start of included source file
				{
					unsigned long int noffset = m_pStabs[m_nStabno].m_nStringOffset;
					std::string file;
					char* cp = m_pStabStrings + noffset;
					if (*cp != '/' || *(cp+1) != '/')
						file = currentpath;
					file += std::string(cp);
					FixPath(file);
					filenames.push(file);
					break;
				}
				case N_EINCL:	// end of included source file
				{
					filenames.pop();
					break;
				}

				case N_SLINE:	// line number
				{
					//  N_SLINE values are relative to function address
					lineaddress = m_pStabs[m_nStabno].m_nValue;
					// ---- add an entry to the source file, line number table
					if (filenames.empty())				
						throw InvalidStab(__LINE__, 0);
					LineAddr lna(nFunctionAddress, lineaddress, filenames.top(), m_pStabs[m_nStabno].m_nDesc);
					int sz = m_SrcLineTable.size();
					if (sz > 0)	{
						LineAddr llna = m_SrcLineTable[--sz];
						if (llna.filename == lna.filename)	{
							if (llna.address == lna.address)	{
								m_SrcLineTable[sz] = lna;
								break;
							}
						}
					}
					m_SrcLineTable.push_back(lna);
					break;
				}
				case N_LBRAC:
				{
					// ---- entering local scope
					MemoryAddress nThisScope = nFunctionAddress + m_pStabs[m_nStabno].m_nValue;
					scopestack.push(nThisScope);
					PostEnterLocalScope(nThisScope);
					break;
				}
				case N_RBRAC:
				{
					// ---- exiting local scope
					MemoryAddress addr = scopestack.top();
					scopestack.pop();
					PostExitLocalScope(addr, m_pStabs[m_nStabno].m_nValue + nFunctionAddress);
					break;
				}
				case N_GSYM:	// global symbol
					m_pStabs[m_nStabno].m_nValue = GetGlobalAddress(m_pStabStrings + m_pStabs[m_nStabno].m_nStringOffset);
					// --- fall through
				case N_LSYM:	// stack variable or type
				case N_PSYM:	// parameter symbol
				case N_LCSYM:	// static file scope variable
				case N_STSYM:	// static global variable
					BuildSymbol();
					break;

				case N_RSYM:	// register variable: Ignore, gets confused w/ range types
					break;

				case N_FUN:		// function
				{
					BuildSymbol();
					char* cp1 = m_pStabStrings + m_pStabs[m_nStabno].m_nStringOffset;
					if (*cp1 == '\0')	{
						// ----- end of function
						highmem = m_pStabs[m_nStabno].m_nValue + nFunctionAddress;
						m_FunctionExits[nFunctionAddress] = highmem - 1;
						if (scopestack.size() != 0)
							throw InvalidStab(__LINE__, 0);
					}
					else	{
						// ----- function declaration (or const global variable)
						nFunctionAddress = m_pStabs[m_nStabno].m_nValue;
						ASSERT(!filenames.empty());
						nCurrentFunctionScope = nFunctionAddress;
						if (scopestack.size() != 0)
							throw InvalidStab(__LINE__, 0);
					}
					break;
				}
				default:
					break;
			}
		}
		catch (foobar)	{
			continue;
		}
	}
	if (!filenames.empty())
		throw InvalidStab(__LINE__, 0);
	return highmem;
}

// --- fixes a path and file specification as recorded in stabs into a format
//     that works with dos and for comparing specs between the stabs entry 
//     and what is provided from the file open dialogs
//
//     All drive, file and sibdirectory names are converted to lowercase
//     //C/ at the beginning becomes c:\
//     forward slashes (/) are converted to backslashes (\)
//     Dot (.) and dot dot (..) subdirectory specs are eliminated. For example:
//
//       c:\myproject\.\include
//             becomes 
//       c:\myproject\include
//
//       c:\myproject\library\..\source
//             becomes 
//       c:\myproject\source
//

void CStab::FixPath(std::string& path)
{
	char newpath[MAX_PATH];
	const char* cp1 = path.c_str();
	char* cp2 = newpath;
	if (*cp1 == '/' && *(cp1+1) == '/' && *(cp1+3) == '/')	{
		// --- convert "//C/" to "c:\"
		*cp2++ = tolower(*(cp1+2));	// drive letter
		*cp2++ = ':';
		*cp2++ = '\\';
		cp1 += 4;
	}
	std::stack<char*> slashes;
	while (*cp1)	{
		if (*cp1 == '/')	{
			*cp2++ = '\\';
			slashes.push(cp2);
			cp1++;
			while (*cp1 == '.')	{
				cp1++;
				if (!*cp1)
					break;
				if (*cp1 == '.')	{
					// --- \..\  ...
					cp1++;
					if (*cp1)
						cp1++;
					slashes.pop();	// pop the slash just before the ..
					cp2 = slashes.top();
				}
				else
					cp1++;		// bypass \.\  ...
			}
		}
		else	{
			*cp2++ = tolower(*cp1);
			cp1++;
		}
	}
	*cp2 = '\0';
	path = newpath;
}

long CStab::GetGlobalAddress(char* sname)
{
	char* cp2 = sname;
	BypassIdentifier(&cp2);
	*cp2 = '\0';
	std::string nm(sname);
	std::map<std::string, MemoryAddress>::iterator iter = imagesymbols.find(nm);
	if (iter == imagesymbols.end())
		return 0;
	return (*iter).second;
}

void CStab::BuildSymbol()
{
	char* cp1 = m_pStabStrings + m_pStabs[m_nStabno].m_nStringOffset;
	if (*cp1 != '\0')	{
		// ----- something with a name, convert : to \0
		char* cp2 = cp1;
		BypassIdentifier(&cp2);
		*cp2 = '\0';
		IncrStrPtr(&cp2);
		TypeIndex ndx = AccumulateTypes(cp1, &cp2);
		// --- if it has an address put it in the symbol table
		if (m_pStabs[m_nStabno].m_nValue != 0)
			AddSymbol(m_pStabs[m_nStabno], ndx);
	}
}

// -------- add a symbol from a stab entry to the symbol table
void CStab::AddSymbol(StabEntry& rStab, TypeIndex ndx)
{
	Symbol symbol(rStab.m_nValue, ndx);

	// --- identifier
	std::string name(m_pStabStrings + rStab.m_nStringOffset);

	// ---- storage class, scope, and backward references scope addresses
	switch (rStab.m_nType)	{
		case N_LSYM:	// stack variable
			symbol.storageclass = Symbol::stack;
			symbol.scope = Symbol::local;
			break;
		case N_PSYM:	// parameter
			symbol.storageclass = Symbol::stack;
			symbol.scope = Symbol::local;
			break;
		case N_RSYM:	// register variable
			symbol.storageclass = Symbol::registr;
			symbol.scope = Symbol::local;
			break;
		case N_GSYM:	// global
			symbol.storageclass = Symbol::external;
			symbol.scope = Symbol::global;
			symbol.enterscope = m_nGlobalScopeEntry;
			break;
		case N_LCSYM:	// static variable
		case N_STSYM:	// initialized static variable
		case N_FUN:		// const global symbol or function
		{
			symbol.storageclass = Symbol::external;
			char *cp = m_pStabStrings + rStab.m_nStringOffset; 
			cp += strlen(cp) + 1;
			if (*cp == 'S')	{
				symbol.scope = Symbol::file;
				symbol.enterscope = m_nCurrentFileScope;
			}
			else if (*cp == 'F')
				return;
			else
				symbol.scope = Symbol::local;
			break;
		}
		default:
			break;
	}
	// ---- add the symbol to the symbol table
	m_Symbols.insert(std::pair<std::string, Symbol>(name, symbol));
}
// --- post enter local scope address to symbols in symbol table
void CStab::PostEnterLocalScope(MemoryAddress nMemAddr)
{
	std::multimap<std::string, Symbol>::iterator iter;
	for (iter = m_Symbols.begin(); iter != m_Symbols.end(); iter++)	{
		Symbol& rSym = (*iter).second;
		if (rSym.scope == Symbol::local && rSym.enterscope == 0)
			rSym.enterscope = nMemAddr;
	}
}
void CStab::PostExitLocalScope(MemoryAddress nEnterScope, MemoryAddress nExitScope)
{
	std::multimap<std::string, Symbol>::iterator iter;
	for (iter = m_Symbols.begin(); iter != m_Symbols.end(); iter++)	{
		Symbol& rSym = (*iter).second;
		if (rSym.scope == Symbol::local && rSym.exitscope == 0 && rSym.enterscope == nEnterScope)
			rSym.exitscope = nExitScope;
	}
}

// --- post scope exit address to symbols in symbol table
void CStab::PostExitScope(MemoryAddress nMemAddr, Symbol::Scope nScope)
{
	std::multimap<std::string, Symbol>::iterator iter;
	for (iter = m_Symbols.begin(); iter != m_Symbols.end(); iter++)	{
		Symbol& rSym = (*iter).second;
		if (rSym.scope == nScope && rSym.exitscope == 0)
			rSym.exitscope = nMemAddr;
	}
}

const Symbol* CStab::FindSymbol(const CString& strSymbol, MemoryAddress nCurrentPC) const
{
	const Symbol* pSym = 0;
	// ---- strip trailing spaces from the search string
	int ni = strSymbol.GetLength();
	while (strSymbol[ni-1] == ' ')
		--ni;
	std::string name(strSymbol.Left(ni));
TRACE("\nLooking for: %s", name.c_str());
	// ----- search for exact match on symbol
	std::multimap<std::string, Symbol>::const_iterator iter = m_Symbols.find(name);
	if (iter != m_Symbols.end())	{
TRACE("\nFound: %s", name.c_str());
		// ---- found a hit on symbol
		if (nCurrentPC != 0)	{
			// ---- iterate forward to last symbol with the same name
			do
				iter++;
			while (iter != m_Symbols.end() && (*iter).first == name);
			--iter;
			// ---- iterate backward to symbol with scope that encloses current program counter
			while ((*iter).first == name)	{
				if ((*iter).second.enterscope <= nCurrentPC && (*iter).second.exitscope > nCurrentPC)	{
					pSym = &(*iter).second;
					break;
				}
				if (iter == m_Symbols.begin())
					break;
				--iter;
			}
		}
	}
	return pSym;
}

// ---- return the memory address associated with a source file and line number
MemoryAddress CStab::SourceLineAddress(const CString& strFile, int lineno)
{
	std::vector<LineAddr>::const_iterator iter;
	for (iter = m_SrcLineTable.begin(); iter != m_SrcLineTable.end(); iter++)	{
		CString nm((*iter).filename.c_str());
		if (strFile.Right(nm.GetLength()).CompareNoCase(nm) == 0 && (*iter).lineno == lineno)
			break;
	}
	return iter != m_SrcLineTable.end() ? ((*iter).funcaddress + (*iter).address) : 0;
}
// ----- return the highest line number in a source code file
int CStab::HighestFileLineNumber(const CString& fname) const
{
	int lineno = 0;
	std::vector<LineAddr>::const_iterator iter;
	for (iter = m_SrcLineTable.begin(); iter != m_SrcLineTable.end(); iter++)	{
		CString nm((*iter).filename.c_str());
		if (fname.CompareNoCase(nm) == 0)
			lineno = (*iter).lineno;
		else if (lineno != 0)
			break;
	}
	return lineno;
}

MemoryAddress CStab::FunctionExitAddress(MemoryAddress addr) const
{
	std::map<MemoryAddress, MemoryAddress>::const_iterator iter;
	iter = m_FunctionExits.lower_bound(addr);
	if (iter == m_FunctionExits.begin())
		return 0;
	return (*--iter).second;
}

// ------- derive the closest source file and line number from a memory address
void CStab::SourceFileAndLine(MemoryAddress memaddr, CString* pstrFile, int* pLineNo)
{
	// ---- binary search algorithm
	long diff; 
	int mid = -1;
	int lo = 0;
	int hi = m_SrcLineTable.size();
	while (lo <= hi)   {
		mid = (lo + hi) / 2;
		diff = memaddr;
		diff -= m_SrcLineTable[mid].funcaddress + m_SrcLineTable[mid].address;
		if (diff == 0)
			break;
		if (diff < 0)
			hi = mid-1;
		else if (diff)
			lo = mid + 1;
	}
	if (mid == -1)	{
		*pLineNo = 0;
		*pstrFile = CString();
	}
	else	{
		*pLineNo = m_SrcLineTable[mid].lineno;
		*pstrFile = CString(m_SrcLineTable[mid].filename.c_str());
	}
}

Type* CStab::GetType(TypeIndex typex)
{ 
	return m_Types[typex];
}

//----- test a stab entry for being a structure, union, or class definition
BOOL CStab::isStructure(StabEntry& rStab)
{
	char* cp = m_pStabStrings + rStab.m_nStringOffset;
	cp += strlen(cp) + 1;
	if (*cp == 'T' || *cp == 't')	{
		IncrStrPtr(&cp);
		BypassTypeIndex(&cp);
		if (*cp == '=')	{
			IncrStrPtr(&cp);
			return *cp == 's' || *cp == 'u';
		}
	}
	return FALSE;
}

// ------- test a stab entry to see if it's a continuation of the previous entry
BOOL CStab::isContinuation(StabEntry& rStab)
{
	char* cp = m_pStabStrings + rStab.m_nStringOffset;
	return (cp > m_pStabStrings+1 && *(cp - 2) == '\\');
}



#ifdef _DEBUG

void CStab::DumpIndex(const TypeIndex& ndx)
{
	TRACE("%d:(%d,%d) ", ndx.fileno, ndx.hdrno, ndx.typeno);
}

#define TN(t) #t

template <class T>
T* DumpTypeData(Type* pType)
{
	T* p = dynamic_cast<T*>(pType);
	if (p != 0)	{
		std::string str(typeid(T).name());
		str = str.substr(7, str.length() - 7);
		TRACE("%s: ", str.c_str());
	}
	return p;
}

#define DT(t) t* p##t = DumpTypeData<t>(pType)

void CStab::DumpType(std::map<TypeIndex, Type*>::iterator iter, bool bValues)
{
	TypeIndex ndx = (*iter).first;
	Type* pType = (*iter).second;
	if (bValues)
		TRACE("\n");
	DumpIndex(ndx);

	if (DT(Intrinsic))
		TRACE("%s, width = %d ", pIntrinsic->m_pName, pIntrinsic->m_nWidth);
	else if (DT(Typedef))	{
		TRACE("%s ", pTypedef->m_pName);
		TRACE("typedef of ");
		DumpIndex(pTypedef->m_TypeIndex);
	}
	else if (DT(Reference))	{
		TRACE("Refers to ");
		DumpIndex(pReference->m_TypeIndex);
	}
	else if (DT(Pointer))	{
		TRACE("Points to ");
		DumpIndex(pPointer->m_TypeIndex);
	}
	else if (DT(Struct))	{
		TRACE("%s ", pStruct->m_pName);
		std::vector<Member>::const_iterator iter;
		for (iter = pStruct->members.begin(); bValues && iter != pStruct->members.end(); iter++)	{
			TRACE("\n   Member: %s ", iter->m_pName);
			DumpIndex(iter->m_TypeIndex);
		}
	}
	else if (DT(Array))	{
		TRACE("%d elements of ", pArray->m_nElements);
		DumpIndex(pArray->m_TypeIndex);
	}
	else if (DT(Enumerated))	{
		TRACE("%s ", pEnumerated->m_pName);
		std::vector<EnumValue>::const_iterator iter;
		for (iter = pEnumerated->values.begin(); bValues && iter != pEnumerated->values.end(); iter++)
			TRACE("\n   %s = %d", iter->m_pName, iter->m_nValue);
	}
	else
		TRACE(" Unknown");
}

void CStab::DumpSymbols()
{
	std::multimap<std::string, Symbol>::iterator iter;
	TRACE("\n------------------------ Symbol Table ----------------------------");
	TRACE("\nAddress  Enters   Exits    Storage   Scope");
	TRACE("\n-------- -------- -------- --------- ------");
	for (iter = m_Symbols.begin(); iter != m_Symbols.end(); iter++)	{
		Symbol& rSym = (*iter).second;
		TRACE("\n%p ", rSym.address);
		TRACE("%p %p ",rSym.enterscope, rSym.exitscope);
		switch (rSym.storageclass)	{
			case Symbol::unknown:
				TRACE("unknown   ");
				break;
			case Symbol::stack:					// in the stack frame (local, arguments)
				TRACE("stack     ");
				break;
			case Symbol::registr:				// register variable
				TRACE("register  ");
				break;
			case Symbol::external:				// globals and statics
				TRACE("external  ");
				break;
			case Symbol::function:				// symbol names a function
				TRACE("function  ");
				break;
			case Symbol::structure:				// symbol names a structure, union, or class
				TRACE("structure ");
				break;
			default:
				TRACE("invalid   ");
				break;
		}
		switch (rSym.scope)	{
			case Symbol::global:
				TRACE("global ");
				break;
			case Symbol::file:
				TRACE("file   ");
				break;
			case Symbol::local:
				TRACE("local  ");
				break;
			default:
				TRACE("invalid   ");
				break;
		}
		DumpIndex(rSym.m_TypeIndex);
		TRACE("%s", (*iter).first.c_str());
	}
	TRACE("\n-----------------------------------------------------------\n");
}

void CStab::DumpSourceLine()
{
	std::vector<LineAddr>::const_iterator iter;
	for (iter = m_SrcLineTable.begin(); iter != m_SrcLineTable.end(); iter++)
		TRACE("\n%08x %0.3d %s", (*iter).funcaddress + (*iter).address, (*iter).lineno, (*iter).filename.c_str());
}

void CStab::DumpTypes()
{
	TRACE("\n------------------------");
	std::map<TypeIndex, Type*>::iterator iter;
	for (iter = m_Types.begin(); iter != m_Types.end(); iter++)
		DumpType(iter);
	TRACE("\n------------------------\n");
}
#endif	// _DEBUG

