// ------- stabs.h

#ifndef STABS_H
#define STABS_H

#include <iostream>
#include <map>
#include <vector>
#include <stack>
#include <set>
#include <string>
#include <winnt.h>

// ------ stab type values
const unsigned char N_GSYM   = 0x20;	//    Global symbol
const unsigned char N_FUN    = 0x24;	//    Function name
const unsigned char N_STSYM  = 0x26;	//    Data segment file-scope variable
const unsigned char N_LCSYM  = 0x28;	//    BSS segment file-scope variable
const unsigned char N_RSYM   = 0x40;	//    Register variable
const unsigned char N_SLINE  = 0x44;	//	  Source code line number
const unsigned char N_EHDECL = 0x50;	//    GNU C++ exception variable
const unsigned char N_CATCH  = 0x54;	//    GNU C++ `catch' clause
const unsigned char N_SSYM   = 0x60;	//    Structure of union element
const unsigned char N_SO     = 0x64;	//    Path and name of source file
const unsigned char N_LSYM   = 0x80;	//    Stack variable or type
const unsigned char N_BINCL  = 0x82;	//    Begin of an include file
const unsigned char N_SOL    = 0x84;	//    Name of include file
const unsigned char N_PSYM   = 0xa0;	//    Parameter variable
const unsigned char N_EINCL  = 0xa2;	//    End of an include file
const unsigned char N_LBRAC  = 0xc0;	//    Beginning of a lexical block
const unsigned char N_EXCL	 = 0xc2;	//    Deleted include file
const unsigned char N_RBRAC  = 0xe0;	//    End of a lexical block

class InvalidStab {
public:
	int err;
	char* cp;
	InvalidStab(int er, char *p) : err(er), cp(p) { }
};

// ------ type index: used to identify types
struct TypeIndex	{
	int fileno;			// the file number where the type is declared
	int hdrno;			// the header number within the file (0 = source file, not header)
	int typeno;			// the number of the type within the file/header
	TypeIndex(int fno = 0, int hno = 0, int tno = 0);
	bool operator<(const TypeIndex& x) const;
	bool operator==(const TypeIndex& x) const;
	bool operator!=(const TypeIndex& x) const
		{ return !(*this == x); }
};

// ------ data types
struct Type	{
	virtual const char* GetName() const { return 0; }
	virtual int GetWidth() const { return 0; }
	virtual ~Type() { }
};

struct Typedef : public Type	{
	char* m_pName;				// name of typedef
	TypeIndex m_TypeIndex;		// index into type table of actual type
	Typedef(char* nm = 0, TypeIndex ndx = TypeIndex()) : m_pName(nm), m_TypeIndex(ndx) { }
	const char* GetName() const { return m_pName; }
};

struct Intrinsic : public Type	{
	char* m_pName;					// name of instrinsic type
	short int m_nWidth;				// number of bytes in the type
	Intrinsic(char* nm, short int wd) : m_pName(nm), m_nWidth(wd) { }
	const char* GetName() const { return m_pName; }
	int GetWidth() const { return m_nWidth; }
};

struct Reference : public Type	{
	TypeIndex m_TypeIndex;		// index into type table of type referred to
	Reference(TypeIndex ndx) : m_TypeIndex(ndx) { }
};

struct Pointer : public Type	{
	TypeIndex m_TypeIndex;		// index into type table of type pointed to
	Pointer(TypeIndex ndx) : m_TypeIndex(ndx) { }
};

// ------ members of classes, structs, unions
struct Member	{
	char* m_pName;				// name of member
	TypeIndex m_TypeIndex;		// index into type table
	short int m_nOffset;		// byte offset from start of struct
	Member(char* nm = 0, TypeIndex ndx = TypeIndex(), short int offset = 0) : 
			m_pName(nm), m_TypeIndex(ndx), m_nOffset(offset) { }
	// --- so Member objects can go in a vector
	bool operator<(const Member& x) const
		{ return m_nOffset < x.m_nOffset; }
	bool operator==(const Member& x) const
		{ return m_nOffset == x.m_nOffset; }
	const char* GetName() const { return m_pName; }
};

struct Struct : public Type	{
	char* m_pName;					// name of class, struct, or union
	short int m_nWidth;				// number of bytes in the type
	std::vector<Member> members;
	Struct(char* nm = 0) : m_pName(nm), m_nWidth(0) { }
	const char* GetName() const { return m_pName; }
	int GetWidth() const { return m_nWidth; }
};

struct Array : public Type	{
	TypeIndex m_TypeIndex;		// index into type table
	short int m_nElements;		// if array, # of elements, else zero
	Array(TypeIndex ndx, short int el) : m_TypeIndex(ndx), m_nElements(el) { }
};

struct EnumValue	{
	char* m_pName;				// name assigned to value
	int m_nValue;				// value assigned to name
	EnumValue(char* nm = 0, int val = 0) : m_pName(nm), m_nValue(val) { }
	// --- so EnumValue objects can go in a vector
	bool operator<(const EnumValue& x) const
		{ return m_nValue < x.m_nValue; }
	bool operator==(const EnumValue& x) const
		{ return m_nValue == x.m_nValue; }
};

struct Enumerated : public Type	{
	char* m_pName;					// name of enum
	std::vector<EnumValue> values;
	Enumerated(char* nm) : m_pName(nm) { }
	const char* GetName() const { return m_pName; }
	int GetWidth() const { return 1; }
};

// ------ symbol table entry
struct Symbol	{
	enum StorageClass {
		unknown,
		stack,					// in the stack frame (local, arguments)
		registr,				// a register variable
		external,				// globals and statics
		function,				// symbol names a function
		structure				// symbol names a structure, union, or class
	} storageclass;				// defines how the address is applied
	enum Scope {
		global,
		file,
		local
	} scope;					// symbol scope
	MemoryAddress address;		// symbol's address
	MemoryAddress enterscope;	// code address where symbol enters scope
	MemoryAddress exitscope;	// code address where symbol leaves scope
	TypeIndex m_TypeIndex;		// index into type table
	Symbol(MemoryAddress addr = 0, TypeIndex ndx = TypeIndex()) : 
			storageclass(unknown), scope(global), address(addr), m_TypeIndex(ndx), enterscope(0), exitscope(0)
		{ }
};

union TypeUnion {
	int					T1;
	char				T2;
	long int			T3;
	unsigned int		T4;
	long unsigned int	T5;
	short int			T8;
	short unsigned int	T9;
	signed char			T10;
	unsigned char		T11;
	float				T12;
	double				T13;
	long double			T14;
	void*				T15;
};

struct StabEntry {
	unsigned long int m_nStringOffset;
	unsigned char m_nType;
	unsigned char m_nOther;
	unsigned short int m_nDesc;
	MemoryAddress m_nValue;
};

struct LineAddr	{
	MemoryAddress funcaddress;	// base address of function
	MemoryAddress address;		// relative program counter of line number
	std::string filename;		// source file
	int lineno;					// line number

	LineAddr(MemoryAddress fadr = 0, MemoryAddress adr = 0, const std::string& fn = std::string(), int lno = 0) : 
						funcaddress(fadr), address(adr), filename(fn), lineno(lno) { }
	// --- so LineAddr objects can go in a vector
	bool operator<(const LineAddr& x) const
		{ return address+funcaddress < x.address+x.funcaddress; }
	bool operator==(const LineAddr& x) const
		{ return address+funcaddress == x.address+x.funcaddress; }
};

class CStab	: public CObject {
	Debugger* m_pDebugger;				// -> owning debugger
	char* m_pStabStrings;				// -> stab strings
	StabEntry* m_pStabs;				// -> stabs

	int m_nFileno;						// source file number
	int m_nNbrStabs;					// number of stabs
	int m_nStabno;						// running stab number
	std::vector<LineAddr> m_SrcLineTable;
	static int nIntrinsicTypeWidths[];	// table of intrinsic data type widths

	std::map<TypeIndex, Type*> m_Types;
	std::multimap<std::string,Symbol> m_Symbols; 

	std::map<std::string, MemoryAddress> imagesymbols;
	std::map<MemoryAddress, MemoryAddress> m_FunctionExits;
	std::stack<MemoryAddress> scopestack;

	MemoryAddress m_nCurrentFileScope;	// current file scope entry address
	MemoryAddress m_nGlobalScopeEntry;	// global scope entry address

	void DeleteTables();
	void InitializeTables();
	int IterateStabs(int nBegin, int nEnd);
	std::set<int> skippers;
	void AddSymbol(StabEntry& rStab, TypeIndex ndx);
	void IncrStrPtr(char** cp, int n = 1);
	void PostExitScope(MemoryAddress nMemAddr, Symbol::Scope nScope);
	void PostEnterLocalScope(MemoryAddress nMemAddr);
	void PostExitLocalScope(MemoryAddress nEnterScope, MemoryAddress nExitScope);
	void BypassDigits(char** cp);
	void BypassNonDigits(char** cp);
	void BypassSpaces(char** cp);
	void FixPath(std::string& path);
	void BypassTypeIndex(char** cp);

	TypeIndex ExtractGenericType(char** cp) throw(InvalidStab);

	TypeIndex AccumulateTypes(char* pName, char** cp);
	TypeIndex AccumulateType(char* pName, char** cp, TypeIndex typex = TypeIndex()) throw(InvalidStab);
	TypeIndex AccumulateIntrinsicType(char* pName, char** cp);
	TypeIndex AccumulateArrayType(char** cp, TypeIndex typex) throw(InvalidStab);
	TypeIndex AccumulateParameterType(char* pName, char** cp);
	TypeIndex AccumulateReferenceType(char** cp, TypeIndex typex) throw (InvalidStab);
	TypeIndex AccumulatePointerType(char** cp, TypeIndex typex) throw (InvalidStab);
	TypeIndex AccumulateRangeType(char** cp);
	TypeIndex AccumulateEnumType(char** cp, TypeIndex typex) throw(InvalidStab);
	TypeIndex AccumulateStructType(char** cp, TypeIndex typex) throw (InvalidStab);
	TypeIndex AccumulateMemberFunction(char** cp, TypeIndex typex) throw (InvalidStab);

	BOOL isIdentifierCharacter(char ch);
	void BypassIdentifier(char** cp);
	void SkipToEndOfMembers(char** cp);
	BOOL isStructure(StabEntry& rStab);
	BOOL isContinuation(StabEntry& rStab);
	void NeedCharacter(char** cp, char ch, int lineno) throw (InvalidStab)
	{
		if (**cp != ch)
			throw(InvalidStab(lineno, *cp));
	}
	void BuildSymbol();
	long GetGlobalAddress(char* sname);

#ifdef _DEBUG
	void DumpSymbols();
	void DumpTypes();
	void DumpSourceLine();
	void DumpIndex(const TypeIndex& ndx);
	void DumpType(std::map<TypeIndex, Type*>::iterator iter, bool bValues = true);
#endif

public:
	CStab(Debugger* pDebugger);
	~CStab();
	BOOL LoadStabs(const CString& strCmd) throw (InvalidStab);
	MemoryAddress SourceLineAddress(const CString& strFile, int lineno);
	void SourceFileAndLine(MemoryAddress memaddr, CString* pstrFile, int* pLineNo);
	Type* GetType(TypeIndex typex);
	const Symbol* FindSymbol(const CString& strSymbol, MemoryAddress nCurrentPC = 0) const;
	int StabLineNo() const
		{ return m_nStabno + 1; }
	const std::vector<LineAddr>& GetLineAddressTable() const
		{ return m_SrcLineTable; }
	int HighestFileLineNumber(const CString& fname) const;
	MemoryAddress FunctionExitAddress(MemoryAddress addr) const;
};

#endif
