// ------ debugger.h

#ifndef DEBUGGER_H
#define DEBUGGER_H


#include "breakpoint.h"

class CStab;

class Debugger	{
	BOOL m_bProcessRunning;		// true while application is running
	BOOL m_bStepping;			// true if executing program is stepping
	BOOL m_bSteppingOver;		// true if executing program is stepping over a function call
	BOOL m_bSteppingOut;		// true if executing program is stepping out of function
	BOOL m_bStopping;			// true while stopped waiting for acknowledgement
	BOOL m_bInProgram;			// true when the application program is running
	BOOL m_bTerminated;			// true when application is terminated by Stop command
	BOOL m_bStepCompleted;		// true when a step has been completed
	BOOL m_bAccessViolation;	// true after debugged program has access violation
	BOOL m_bIsBroken;			// true when breakpoint has occurred

	PROCESS_INFORMATION m_ProcessInformation;	// for executing application
	CONTEXT m_Context;			// context of debugged thread
	HWND m_hApplWnd;			// console window handle of executing application
	CString m_strApplTitle;		// window title of executing application
	CString m_strCmd;			// executing command line

	HANDLE m_hProcess;			// handle of debugged application
	HANDLE m_hThread;			// handle of debugged thread
	HANDLE m_hFile;				// handle of target program's exe file
	DWORD m_dwProcessId;		// process id of debugged application
	DWORD m_dwThreadId;			// thread id of debugged application

	MemoryAddress m_nCurrentPC;	// current (virtual) program counter at time of breakpoint or single step
	MemoryAddress m_nBrokenPC;
	MemoryAddress m_nSteppingOutPC;
	unsigned char m_cSteppingOutInst;			// op code of instruction to be broken

	CString m_strCurrentSrcFile;// source file being debugged
	int m_nCurrentLineNo;		// source file line number of current program counter

	CString m_strSteppingSrcFile;	// source code file being stepped in
	int m_nSteppingLineNo;			// line number of step

	char* m_pOldPath;			// original working logged-on subdirectory
	int m_nDrive;				// original working drive

	// ------ debug info (GNU symbol tables, line numbers, etc.)
	CStab* m_pCStabs;

	// ------ runtime breakpoint table
	RuntimeBreakpoints breakpoints;
	static BOOL m_bSteppingTo;	// true if executing program is stepping to the cursor line
	static FileReference steptobreakpoint;	// for stepping to and over source lines
public:
	enum DebugEvent { close, suspend, resume, active, breaking, finished };
private:
	HANDLE m_DebugEvents[6];

	static Debugger* pDebugger;

	int m_nBoolType;		// type # of bool type (if defined, else 0)
	int m_nVoidType;		// type # of void type

private:
	static UINT DebugProg(void*);
	void DebugApplicationProgram();
	void DebugProgram();
	void SetConsoleWindow();
	void RunDebugger();
	void PostTerminationMessage();
	void InitStabs(const CString& strCmd);
	void ShutdownProcess();
	static BOOL CALLBACK EnumWindowsProc(HWND hwnd, LPARAM);
	void LoadAllBreakpoints();

	void CloseDebuggerHandle(HANDLE& hHandle);
public:
	Debugger();
	~Debugger();
	void Execute(CString& strCmd);
	void Stop();
	void ContinueDebugging();
	void ResumeProgram();
	void SetStepping();
	void StopStepping();
	BOOL isStepping() const;
	BOOL isBroken() const;
	CStab* GetStabs() const;
	const CONTEXT& Context() const;
	int CurrentLineNo() const;
	const CString* CurrentSrcFile() const;
	MemoryAddress CurrentPC() const;
	BOOL ProcessRunning() const;
	HANDLE ProcessHandle() const;
	void ShowPCCaret(BOOL bOnOff);
	BOOL CanStop() const;
	BOOL ExecuteProgram();
	BOOL StepProgram();
	BOOL StepOver();
	BOOL StepOut();
	static void StepTo(const CString& strFile, int nLineNo);
	BOOL OKToExamine() const;
	BOOL IsInProgram() const;
	void SelectSteppingSourceLine();
};

inline void Debugger::SetStepping()
{
	m_bStepping = TRUE;
}
inline void Debugger::StopStepping()
{
	m_bStepping = FALSE;
}
inline BOOL Debugger::isStepping() const
{
	return m_bStepping;
}
inline BOOL Debugger::isBroken() const
{
	return m_bIsBroken;
}
inline const CString* Debugger::CurrentSrcFile() const
{
	return &m_strCurrentSrcFile;
}
inline int Debugger::CurrentLineNo() const
{
	return m_nCurrentLineNo;
}
inline CStab* Debugger::GetStabs() const
{
	return m_pCStabs;
}
inline const CONTEXT& Debugger::Context() const
{
	return m_Context;
}
inline MemoryAddress Debugger::CurrentPC() const
{
	return m_nCurrentPC;
}
inline BOOL Debugger::IsInProgram() const
{
	return m_bInProgram;
}
inline BOOL Debugger::ProcessRunning() const
{
	return m_bProcessRunning;
}
inline HANDLE Debugger::ProcessHandle() const
{
	return m_hProcess;
}
inline BOOL Debugger::CanStop() const
{
	return ProcessRunning() && !m_bStopping;
}
inline BOOL Debugger::OKToExamine() const
{
	return !IsInProgram() && (isBroken() || isStepping());
}
#endif


