// -------- parser.h

#ifndef PARSER_H
#define PARSER_H

#include "debugger.h"
#include "stabs.h"

// ----- exception classes
class ParserError {
	CString m_strErrorMsg;
public:
	ParserError(const CString& ermsg = CString()) : m_strErrorMsg(ermsg) { }
	void DisplayMessage(const char* pExpression);
};
class UnknownSymbol : public ParserError { };
class UnknownType : public ParserError { 
public:	
	UnknownType() : ParserError("Unknown data type") { }
};
class LValueError : public ParserError { 
public:	
	LValueError() : ParserError("lvalue required") { }
};
class UnsupportedTypeError : public ParserError { 
public:	
	UnsupportedTypeError() : ParserError("unsupported nonstandard GNU type") { }
};
class PtrArrayError : public ParserError { 
public:	
	PtrArrayError() : ParserError("Pointer or array required") { }
};
class IntegerError : public ParserError { 
public:	
	IntegerError() : ParserError("Integer required") { }
};
class VoidPointerError : public ParserError { 
public:	
	VoidPointerError() : ParserError("Cannot dereference void pointer") { }
};
class PointerAdditionError : public ParserError { 
public:	
	PointerAdditionError() : ParserError("Cannot add 2 pointers") { }
};
class IncompatiblePointerSubtractionError : public ParserError { 
public:	
	IncompatiblePointerSubtractionError() : ParserError("Incompatible pointer subtraction") { }
};
class PointerSubtractionError : public ParserError { 
public:	
	PointerSubtractionError() : ParserError("Can't subtract pointer from non-pointer") { }
};
class NeedRBracketError : public ParserError { 
public:	
	NeedRBracketError() : ParserError("Missing ]") { }
};
class NotStructureError : public ParserError { 
public:	
	NotStructureError() : ParserError("Structure needed on left of . or ->") { }
};
class MemberError : public ParserError { 
public:	
	MemberError() : ParserError("Structure member needed on right of . or ->") { }
};
class NeedRParenError : public ParserError { 
public:	
	NeedRParenError() : ParserError("Missing )") { }
};
class SyntaxError : public ParserError { 
public:	
	SyntaxError() : ParserError("Syntax error") { }
};

union Datum {
	char m_char;
	int m_int;
	unsigned m_unsigned;
	short int m_short;
	unsigned short int m_unsignedshort;
	long int m_long;
	unsigned long int m_unsignedlong;
	float m_float;
	double m_double;
	long double m_longdouble;
	char* m_ptr;
};

class Parser;
struct Type;

class Item {
public:
	enum ItemType { 
		Void, 
		Bool,			// keep this as first integer type 
		Char,
		UnsignedChar,	// keep all integer types together
		Int,
		UnsignedInt,
		Short, 
		UnsignedShort,
		Long, 
		UnsignedLong,
		Enum, 			// keep this as last integer type
		Float, 
		Double, 
		LongDouble, 
		Structure 
	};
private:
	enum ItemType m_Type;
    Datum m_Datum;					// data value or address
	CString m_strTypeDisp;			// -> display of type for watches, examines
	short int m_nIndirection;		// levels of indirection (0 = not a pointer)
	short int m_nDimensions;		// number of dimensions in an array (0 = not an array)
	short int* m_pDimensions;		// -> array of dimensions
	short int m_nWidth;				// width of one item or, if array, one element
	MemoryAddress m_nMemaddr;		// address of item
	TypeIndex m_StabType;			// index into stab type table
	Type* m_pType;					// -> type data from stabs
	short int m_nReferences;		// for displaying reference type

	void NullItem();
	void Copy(Item& rItem);
	void InsertTypeData();
public:
	Item();
	Item(Item& rItem);
	Item(const CString& strIdentifier);
	Item(int nConstant);
	~Item();
	void AddressOf();
	BOOL isPointer() const;
	BOOL isArray() const;
	BOOL isLvalue() const;
	BOOL isPointerOrArray() const;
	void NeedInteger();
	short int Width();
	void GetFromMemory();
	void GetMember(const CString& rstrIdentifier);
	void SetType();
	void DereferencePointer();
	void NoVoidPointer();
	void IntegersOnly(Item& rItem);
	void BuildDisplay();
	void BuildValueDisplay(CString& strVarValue);
	void ChangeMemoryVariable(HANDLE hProcess, const CString& strVarValue);
	Item& operator=(Item& rItem);
	void operator+=(Item& rItem);
	void operator-=(Item& rItem);
	void operator*=(Item& rItem);
	void operator/=(Item& rItem);
	ItemType GetItemType() const
		{ return m_Type; }
	short int GetShort() const
		{ return (short int) m_Datum.m_short; }
	long int GetLong() const
		{ return (long int) m_Datum.m_long; }
	char* GetPtr() const
		{ return m_Datum.m_ptr; }
};

inline BOOL Item::isPointer() const
{
	return m_nIndirection != 0;
}
inline BOOL Item::isArray() const
{
	return m_nDimensions != 0;
}
inline BOOL Item::isLvalue() const
{
	return isPointer() || (m_nMemaddr != 0 && !isArray() && m_Type != Structure);
}
inline BOOL Item::isPointerOrArray() const
{
	return isPointer() || isArray();
}
inline void Item::NeedInteger()
{
	if (m_Type < Bool || m_Type > Enum)
		throw IntegerError();
}

class Parser	{
	char m_Token;
	char* m_pTokenPtr;
	char* m_pTokenBuffer;
	CString m_strIdentifier;
	int m_nConstant;

	void LexicalScanner(const char *pSrcbuf);
	void Constant(const char **srcstr, char **tknstr);
	void NextToken();
	Item AddSubtract();
	Item MultiplyDivide();
	Item Primary();
	Item Element();
public:
	Parser();
	~Parser();
	Item EvaluateExpression(const char *pExpression);
};

#endif
