// ----- multikeymap.h

#ifndef MULTIKEY_MAP_H
#define MULTIKEY_MAP_H

#include <vector>
#include <map>

namespace DDJCProgrammingColumnMultiKeyMap	{
//
// P = primary key
// S = secondary key
// T = data record being contained and indexed
//
template<class P, class S, class T>
class multikey_map	{
public:
	typedef std::vector<T>::iterator iterator;
	typedef std::vector<T>::const_iterator const_iterator;
	typedef std::vector<T>::reverse_iterator reverse_iterator;
	typedef std::vector<T>::const_reverse_iterator const_reverse_iterator;
	typedef std::vector<T>::reference reference;
	typedef std::vector<T>::const_reference const_reference;
	typedef std::vector<T>::size_type size_type;
private:
	std::vector<T> items;
	std::map<P, iterator> primarymap;
	std::multimap<S, iterator> secondarymap;
public:
	void add(const P& p, const S& s, const T& t)
	{
		iterator pos = items.insert(items.end(), t);
		primarymap[p] = pos;
		secondarymap.insert(std::pair<S, iterator>(s, pos));
	}
	size_type size() const
		{ return items.size(); }

	iterator begin()
		{ return items.begin(); }
	const_iterator begin() const
		{ return const_cast<const_iterator>(items.begin()); }
	iterator end()
		{ return items.end(); }
	const_iterator end() const
		{ return const_cast<const_iterator>(items.end()); }
	reverse_iterator rbegin()
		{ return items.rbegin(); }
	const_reverse_iterator rbegin() const
		{ return (const_reverse_iterator)items.rbegin(); }
	reverse_iterator rend()
		{ return items.rend(); }
	const_reverse_iterator rend() const
		{ return (const_reverse_iterator)items.rend(); }

	iterator find(const P& p)
	{
		std::map<P, iterator>::iterator iter = primarymap.find(p);
		if (iter == primarymap.end())
			return end();
		return (*iter).second;
	}
	const_iterator find(const P& p) const
	{
		return const_cast<iterator>(find(p));
	}
	iterator find_secondary(const S& s)
	{
		std::multimap<S, iterator>::iterator iter = secondarymap.lower_bound(s);
		if (iter != secondarymap.end() && (*iter).first == s)
			return (*iter).second;
		return end();
	}
	const_iterator find_secondary(const S& s) const
	{
		return const_cast<iterator>(find_secondary(s));
	}
	bool empty() const
		{ return items.empty(); }
	void clear()
	{
		items.clear();
		primarymap.clear();
		secondarymap.clear();
	}
	void erase(const P& p, const S& s)
	{
		std::map<P, iterator>::iterator pri_iter;
		std::multimap<S, iterator>::iterator sec_iter;
		if ((pri_iter = primarymap.find(p)) != primarymap.end())	{
			iterator iter = items.erase((*pri_iter).second);
			// --- adjust entries in primary map to reflect deleted vector entry
			for (pri_iter = primarymap.begin(); pri_iter != primarymap.end(); pri_iter++)
				if ((*pri_iter).second >= iter)
					--(*pri_iter).second;
			// --- adjust entries in secondary map to reflect deleted vector entry
			for (sec_iter = secondarymap.begin(); sec_iter != secondarymap.end(); sec_iter++)
				if ((*sec_iter).second >= iter)
					--(*sec_iter).second;
		}
	}
};


} //  namespace CProgrammingColumnMultiKeyMap

#endif
