// ----- song.h

#ifndef SONG_H
#define SONG_H

#include "midicord.h"

const int maxchords = 1024;

class UndoSongData;

#define UNTITLED "(Untitled)"

class Song : public CObject	{
	DECLARE_SERIAL(Song);
public:
	enum Style { 
		nostyle, 
		fourfour, 
		halftime, 
		rock, 
		country,
		waltz, 
		bossa, 
		rockballad, 
		easybossa,
		jazzwaltz, 
		ballad, 
		dixie,
		shuffle
	};
	enum EventType {
		none            =  0,
		style           =  1, 
		beginrepeat     =  2, 
		endrepeat       =  4,
		rest            =  8,
		ending          = 16,
//		beginwithending = 32,
		passage         = 64
	};
	struct Event	{
		CChord m_Chord;
		EventType m_nEventType;
		Style m_nStyleChange;
		int m_nEndingNumber;
		Event() : m_nEventType(none), m_nStyleChange(nostyle), m_nEndingNumber(0)
			{ }
	};
	static CString m_strDefaultTitle;
	static Song::Style m_aButtonMap[];
// --- song data members
public:
	CString m_strTitle;
	CChord m_Key;
	Style m_nStyle;
	int m_nTempo;
	int m_nBeatsPerMeasure;
private:
	unsigned int m_nChordCount;
	static Event m_aEvents[];
	struct StyleText {
		char* m_pszName;
		char* m_pszNotation;
	};
	static StyleText m_Styles[];

	UndoSongData* m_pUndoSongData;

// --- playback data members
public:
	CChord m_Chord;
private:
	int m_nMeasureCount;
	int m_nCurrentChord;
	int m_nNextChord;
	int m_nFirstChord;
	int m_nLastChord;
	int m_nCurrentMeasure;
	BOOL m_bPlayback;
	BOOL m_bWasPlayback;
	BOOL m_bPlaying;
	BOOL m_bRecording;
	BOOL m_bLastMeasure;
	Style m_nNewStyle;
	int m_nBeat;
	int m_nClock;
	int m_nRepeatTo;
	BOOL m_bRepeat;
	BOOL m_bCued;
	int m_nPlayingEnding;
	int m_nRepeatMeasure;
	void TransposeNote(short int& note, short int delta);
	void TransposeChord(CChord& rChord, short int delta);
// ---- public interface
public:
	Song();
	~Song();
	void InitializeSong();
	void BuildDummySong();
	void ResetChordOffset();
	void PostChord(const CChord& rChord);
	void BuildChordDisplay(CString& cstr, const midiChord& Chord, BOOL bAbbrev = false);
	BOOL IsPlayback() const
		{ return m_bPlayback; }
	BOOL IsPlaying() const
		{ return m_bPlaying; }
	void SetPlayback(int nFirstMeasure, int nLastMeasure);
	BOOL IsWaltz() const
		{ return m_nStyle == jazzwaltz || m_nStyle == waltz; }
	void StartPlaying(BOOL bRecording);
	BOOL TimingMessage(int beat, int clock);
	void StopPlaying();
	int ChordCount() const
		{ return (m_bPlaying && (!m_bRecording && !m_bWasPlayback)) ? 0 : m_nChordCount; }
	int CurrentChordNo() const
		{ return m_nCurrentChord; }
	const CChord& GetChord(int nChordNo) const
		{ ASSERT(nChordNo < maxchords); return m_aEvents[nChordNo].m_Chord; }
	const Event& GetEvent(int nEventNo) const
		{ ASSERT(nEventNo < maxchords); return m_aEvents[nEventNo]; }

	void SetTitle(const CString& strTitle);
	void SetStyle(Style nStyle);
	void SetStyle(int nMeasure, Style nStyle);
	void ReplaceChord(int nChordNo, const CChord& rChord, BOOL bStophere);
	void SetEvent(int nEventNo, EventType nType, BOOL bStophere = TRUE);
	void ClearEvent(int nEventNo, EventType nType, BOOL bStophere = TRUE);
	void SetKey(const CChord& Key);
	void SetTempo(int nTempo);
	void Transpose(const CChord& NewKey);

	void Serialize(CArchive& archive);
	const char* StyleNotation(Style nStyle) const
		{ return m_Styles[(int)nStyle].m_pszNotation; }
	const char* StyleName(Style nStyle) const
		{ return m_Styles[(int)nStyle].m_pszName; }
	void Cue()
		{ m_bCued = TRUE; }
	void DeleteMeasures(int nMeasureNo, int nCount, BOOL bStophere = TRUE);
	void InsertMeasures(int nMeasureNo, int nCount, Event* pEvent, BOOL bStophere = TRUE);
	void DeleteChord(unsigned int nChordNo);
	// --- for undo's use
	void Insert(Event* pInsertAt, int nCount, Event* pEvent);
	void Delete(Event* pDeleteAt, int nCount);
	void ReplaceTitle(CString strTitle);
	void ReplaceTranspose(short int nDelta);

	void UndoLastAction();
	BOOL IsUndoDataStored() const;
	void ResetChanged();
	BOOL IsCued() const
		{ return m_bCued; }
	BOOL CanSave() const
		{ return IsUndoDataStored() || m_strTitle == UNTITLED; }
	BOOL IsChanged() const
		{ return IsUndoDataStored(); }
	BOOL IsRest(int nChord) const
		{ ASSERT(nChord < maxchords); return (m_aEvents[nChord].m_nEventType & rest) != 0; }
	BOOL IsRest(int nMeasure, int nBeat) const
		{ return IsRest(nMeasure * m_nBeatsPerMeasure + nBeat); }
	BOOL IsBeginRepeat(int nChord) const
		{ ASSERT(nChord < maxchords); return (m_aEvents[nChord].m_nEventType & beginrepeat) != 0; }
//	BOOL IsBeginRepeatWithEnding(int nChord) const
//		{ ASSERT(nChord < maxchords); return (m_aEvents[nChord].m_nEventType & beginwithending) != 0; }
	BOOL IsEndRepeat(int nChord) const
		{ ASSERT(nChord < maxchords); return (m_aEvents[nChord].m_nEventType & endrepeat) != 0; }
	BOOL IsLastMeasure() const
		{ return m_bLastMeasure; }
	BOOL IsNextToLastMeasure() const
		{ return m_nCurrentMeasure == m_nMeasureCount - 2; }
	int ButtonFromStyle(Style nStyle) const;
	Style StyleFromButton(int nButton) const
		{ return m_aButtonMap[nButton]; }
	Style CurrentStyle(int nMeasure) const;
	void PadLastMeasure();
	void SetEndingNumber(int nEvent, int nEnding)
		{ ASSERT(nEvent < maxchords); m_aEvents[nEvent].m_nEndingNumber = nEnding; }
	BOOL LoadSong(const CString& strTitle);
#ifndef DEMO_VERSION
	void SaveSong();
#endif
	BOOL Validate(unsigned int& nEvent, CString& strErrMsg);
	int MeasureCount() const
		{ return m_nMeasureCount; }
#ifdef _DEBUG
	void DumpSong() const;
#endif
};

extern Song theSong;

#endif
